"""
Common Sense Reasoning: Physics, social, and temporal reasoning.

Implements intuitive reasoning about the physical and social world.
"""

from dataclasses import dataclass
from typing import Dict, List, Optional, Any, Tuple, Set
from enum import Enum


class PhysicsProperty(Enum):
    """Physical properties of objects."""

    SOLID = "solid"
    LIQUID = "liquid"
    GAS = "gas"
    HEAVY = "heavy"
    LIGHT = "light"
    BREAKABLE = "breakable"
    FLEXIBLE = "flexible"
    HOT = "hot"
    COLD = "cold"
    SHARP = "sharp"
    SOFT = "soft"


class SocialRelation(Enum):
    """Social relations between agents."""

    FRIEND = "friend"
    FAMILY = "family"
    COLLEAGUE = "colleague"
    STRANGER = "stranger"
    AUTHORITY = "authority"
    SUBORDINATE = "subordinate"


@dataclass
class PhysicsRule:
    """A physics rule for reasoning."""

    name: str
    condition: str  # Simplified condition description
    effect: str  # Simplified effect description
    confidence: float = 1.0


@dataclass
class SocialNorm:
    """A social norm or expectation."""

    name: str
    context: str
    expected_behavior: str
    violation_consequence: str
    strength: float = 1.0  # How strong is this norm


@dataclass
class TemporalRelation:
    """A temporal relation between events."""

    event1: str
    event2: str
    relation: str  # before, after, during, overlaps, meets
    duration: Optional[float] = None


class CommonSenseKB:
    """
    Common sense knowledge base.

    Stores general world knowledge that humans take for granted.
    """

    def __init__(self):
        # Object properties
        self._object_properties: Dict[str, Set[PhysicsProperty]] = {}

        # Default properties by category
        self._category_defaults: Dict[str, Set[PhysicsProperty]] = {
            "glass": {PhysicsProperty.SOLID, PhysicsProperty.BREAKABLE},
            "water": {PhysicsProperty.LIQUID, PhysicsProperty.COLD},
            "metal": {PhysicsProperty.SOLID, PhysicsProperty.HEAVY},
            "wood": {PhysicsProperty.SOLID},
            "fire": {PhysicsProperty.HOT},
            "ice": {PhysicsProperty.SOLID, PhysicsProperty.COLD},
            "rock": {PhysicsProperty.SOLID, PhysicsProperty.HEAVY},
            "feather": {PhysicsProperty.LIGHT, PhysicsProperty.SOFT},
            "knife": {PhysicsProperty.SOLID, PhysicsProperty.SHARP},
        }

        # Typical object locations
        self._typical_locations: Dict[str, List[str]] = {
            "kitchen": ["stove", "refrigerator", "sink", "knife", "plate"],
            "bedroom": ["bed", "pillow", "lamp", "closet"],
            "bathroom": ["toilet", "shower", "sink", "mirror"],
            "office": ["desk", "computer", "chair", "printer"],
            "garage": ["car", "tools", "workbench"],
        }

        # Typical actions for objects
        self._typical_actions: Dict[str, List[str]] = {
            "knife": ["cut", "slice", "chop"],
            "cup": ["drink", "fill", "pour"],
            "book": ["read", "open", "close"],
            "door": ["open", "close", "lock"],
            "car": ["drive", "park", "start"],
            "phone": ["call", "text", "charge"],
        }

        # Social roles
        self._social_roles: Dict[str, Dict[str, Any]] = {
            "doctor": {"authority": True, "trust": 0.9, "expertise": ["health"]},
            "teacher": {"authority": True, "trust": 0.8, "expertise": ["education"]},
            "friend": {"authority": False, "trust": 0.7, "expertise": []},
            "stranger": {"authority": False, "trust": 0.2, "expertise": []},
        }

    def get_properties(self, obj: str) -> Set[PhysicsProperty]:
        """Get properties of an object."""
        if obj in self._object_properties:
            return self._object_properties[obj]

        # Check category defaults
        for category, props in self._category_defaults.items():
            if category in obj.lower():
                return props

        return set()

    def set_properties(
        self,
        obj: str,
        properties: Set[PhysicsProperty],
    ) -> None:
        """Set properties for an object."""
        self._object_properties[obj] = properties

    def get_typical_location(self, obj: str) -> Optional[str]:
        """Get typical location for an object."""
        for location, objects in self._typical_locations.items():
            if obj in objects:
                return location
        return None

    def get_objects_in_location(self, location: str) -> List[str]:
        """Get typical objects in a location."""
        return self._typical_locations.get(location, [])

    def get_typical_actions(self, obj: str) -> List[str]:
        """Get typical actions for an object."""
        return self._typical_actions.get(obj, [])

    def get_social_role(self, role: str) -> Dict[str, Any]:
        """Get information about a social role."""
        return self._social_roles.get(role, {})


class PhysicsReasoner:
    """
    Intuitive physics reasoning.

    Reasons about physical properties, causation, and mechanics.
    """

    def __init__(self, kb: Optional[CommonSenseKB] = None):
        self.kb = kb or CommonSenseKB()

        # Physics rules
        self._rules: List[PhysicsRule] = [
            PhysicsRule(
                "gravity",
                "object is unsupported",
                "object falls down",
            ),
            PhysicsRule(
                "breakable",
                "breakable object falls on hard surface",
                "object breaks",
            ),
            PhysicsRule(
                "liquid_flow",
                "container with liquid is tilted",
                "liquid flows out",
            ),
            PhysicsRule(
                "heat_transfer",
                "hot object contacts cold object",
                "heat transfers from hot to cold",
            ),
            PhysicsRule(
                "melting",
                "ice is exposed to heat",
                "ice melts into water",
            ),
            PhysicsRule(
                "freezing",
                "water is exposed to cold below freezing",
                "water freezes into ice",
            ),
            PhysicsRule(
                "burning",
                "flammable object contacts fire",
                "object burns",
            ),
            PhysicsRule(
                "floating",
                "light object is in water",
                "object floats",
            ),
            PhysicsRule(
                "sinking",
                "heavy object is in water",
                "object sinks",
            ),
            PhysicsRule(
                "cutting",
                "sharp object contacts soft object with force",
                "soft object is cut",
            ),
        ]

    def predict_effect(
        self,
        obj: str,
        situation: str,
    ) -> List[Tuple[str, float]]:
        """Predict effects of a situation on an object."""
        effects = []
        properties = self.kb.get_properties(obj)

        for rule in self._rules:
            # Simple keyword matching for demonstration
            condition_match = False

            if "falls" in situation.lower() and "unsupported" in rule.condition:
                condition_match = True
            elif "falls" in situation.lower() and PhysicsProperty.BREAKABLE in properties:
                if "breakable" in rule.condition:
                    condition_match = True
            elif "heat" in situation.lower() and "heat" in rule.condition:
                condition_match = True
            elif "fire" in situation.lower() and "fire" in rule.condition:
                condition_match = True
            elif "water" in situation.lower():
                if PhysicsProperty.HEAVY in properties and "heavy" in rule.condition:
                    condition_match = True
                elif PhysicsProperty.LIGHT in properties and "light" in rule.condition:
                    condition_match = True

            if condition_match:
                effects.append((rule.effect, rule.confidence))

        return effects

    def will_break(self, obj: str, action: str) -> float:
        """Estimate probability of object breaking."""
        properties = self.kb.get_properties(obj)

        if PhysicsProperty.BREAKABLE not in properties:
            return 0.0

        # Risk factors
        risk = 0.5  # Base risk for breakable

        if "drop" in action.lower() or "fall" in action.lower():
            risk += 0.4
        if "throw" in action.lower():
            risk += 0.3
        if "hit" in action.lower() or "strike" in action.lower():
            risk += 0.3

        return min(1.0, risk)

    def can_contain(self, container: str, contained: str) -> bool:
        """Check if container can hold contained object."""
        container_props = self.kb.get_properties(container)
        contained_props = self.kb.get_properties(contained)

        # Liquids need solid containers
        if PhysicsProperty.LIQUID in contained_props:
            return PhysicsProperty.SOLID in container_props

        return True  # Simplified

    def get_applicable_rules(self, situation: str) -> List[PhysicsRule]:
        """Get physics rules applicable to a situation."""
        applicable = []

        for rule in self._rules:
            # Check if rule condition keywords are in situation
            keywords = rule.condition.lower().split()
            if any(kw in situation.lower() for kw in keywords):
                applicable.append(rule)

        return applicable


class SocialReasoner:
    """
    Social reasoning and theory of mind.

    Reasons about social situations, norms, and mental states.
    """

    def __init__(self, kb: Optional[CommonSenseKB] = None):
        self.kb = kb or CommonSenseKB()

        # Social norms
        self._norms: List[SocialNorm] = [
            SocialNorm(
                "greeting",
                "meeting someone",
                "say hello or greet",
                "perceived as rude",
                strength=0.9,
            ),
            SocialNorm(
                "thanking",
                "receiving help or gift",
                "express gratitude",
                "perceived as ungrateful",
                strength=0.95,
            ),
            SocialNorm(
                "personal_space",
                "interacting with strangers",
                "maintain appropriate distance",
                "discomfort and distrust",
                strength=0.8,
            ),
            SocialNorm(
                "turn_taking",
                "conversation",
                "wait for others to finish speaking",
                "perceived as rude or aggressive",
                strength=0.7,
            ),
            SocialNorm(
                "honesty",
                "general interaction",
                "tell the truth",
                "loss of trust if discovered",
                strength=0.85,
            ),
            SocialNorm(
                "reciprocity",
                "receiving favor",
                "return favor when possible",
                "relationship strain",
                strength=0.7,
            ),
        ]

    def get_applicable_norms(self, context: str) -> List[SocialNorm]:
        """Get social norms applicable to a context."""
        applicable = []

        for norm in self._norms:
            if norm.context.lower() in context.lower():
                applicable.append(norm)
            elif any(word in context.lower() for word in norm.context.split()):
                applicable.append(norm)

        return applicable

    def predict_reaction(
        self,
        action: str,
        relationship: SocialRelation,
    ) -> Tuple[str, float]:
        """Predict social reaction to an action."""
        # Check norm violations
        for norm in self._norms:
            # Simple check if action violates norm
            if "help" in action.lower() and "thanking" in norm.name:
                if "thank" not in action.lower():
                    return (norm.violation_consequence, norm.strength)

        # Default positive reaction
        return ("neutral or positive", 0.5)

    def infer_goal(
        self,
        observed_actions: List[str],
    ) -> List[Tuple[str, float]]:
        """Infer likely goals from observed actions."""
        goals = []

        for action in observed_actions:
            action_lower = action.lower()

            if "eat" in action_lower or "cook" in action_lower:
                goals.append(("satisfy hunger", 0.8))
            if "work" in action_lower or "study" in action_lower:
                goals.append(("achieve task", 0.7))
            if "talk" in action_lower or "meet" in action_lower:
                goals.append(("social connection", 0.6))
            if "buy" in action_lower or "shop" in action_lower:
                goals.append(("acquire item", 0.7))
            if "help" in action_lower:
                goals.append(("assist others", 0.7))

        return goals

    def infer_emotion(
        self,
        situation: str,
        outcome: str,
    ) -> List[Tuple[str, float]]:
        """Infer likely emotional state."""
        emotions = []

        if "success" in outcome.lower() or "win" in outcome.lower():
            emotions.append(("happy", 0.8))
            emotions.append(("proud", 0.6))
        elif "fail" in outcome.lower() or "lose" in outcome.lower():
            emotions.append(("sad", 0.7))
            emotions.append(("frustrated", 0.6))
        elif "surprise" in situation.lower():
            emotions.append(("surprised", 0.8))
        elif "threat" in situation.lower() or "danger" in situation.lower():
            emotions.append(("afraid", 0.8))
        elif "unfair" in situation.lower():
            emotions.append(("angry", 0.7))

        return emotions

    def trust_level(
        self,
        source_role: str,
        topic: str,
    ) -> float:
        """Estimate trust level for information from source about topic."""
        role_info = self.kb.get_social_role(source_role)

        if not role_info:
            return 0.3  # Unknown role

        base_trust = role_info.get("trust", 0.5)
        expertise = role_info.get("expertise", [])

        # Boost trust if topic matches expertise
        if any(exp in topic.lower() for exp in expertise):
            base_trust = min(1.0, base_trust + 0.2)

        return base_trust


class TemporalReasoner:
    """
    Temporal reasoning about events and time.

    Reasons about sequences, durations, and temporal relations.
    """

    def __init__(self):
        # Typical durations (in minutes)
        self._typical_durations: Dict[str, Tuple[float, float]] = {
            "eating_meal": (15, 60),
            "cooking": (20, 120),
            "showering": (5, 20),
            "sleeping": (300, 600),  # 5-10 hours
            "working": (240, 600),  # 4-10 hours
            "meeting": (30, 120),
            "phone_call": (2, 30),
            "driving_commute": (15, 90),
            "exercise": (20, 90),
            "reading": (15, 120),
        }

        # Event sequences (typical orderings)
        self._typical_sequences: Dict[str, List[str]] = {
            "morning_routine": ["wake_up", "shower", "dress", "eat_breakfast", "leave_home"],
            "cooking": ["prepare_ingredients", "cook", "serve", "eat"],
            "shopping": ["travel_to_store", "browse", "select_items", "pay", "leave"],
            "meeting": ["arrive", "greet", "discuss", "conclude", "leave"],
        }

    def get_typical_duration(
        self,
        activity: str,
    ) -> Optional[Tuple[float, float]]:
        """Get typical duration range for an activity (min, max in minutes)."""
        for key, duration in self._typical_durations.items():
            if key in activity.lower() or activity.lower() in key:
                return duration
        return None

    def get_typical_sequence(self, activity_type: str) -> List[str]:
        """Get typical sequence of events for an activity type."""
        for key, sequence in self._typical_sequences.items():
            if key in activity_type.lower():
                return sequence
        return []

    def order_events(
        self,
        events: List[str],
        context: Optional[str] = None,
    ) -> List[str]:
        """Order events in likely temporal sequence."""
        if context:
            template = self.get_typical_sequence(context)
            if template:
                # Order events according to template
                ordered = []
                remaining = list(events)

                for step in template:
                    for event in remaining:
                        if step in event.lower() or event.lower() in step:
                            ordered.append(event)
                            remaining.remove(event)
                            break

                ordered.extend(remaining)
                return ordered

        return events  # Return original order if no template

    def infer_relation(
        self,
        event1: str,
        event2: str,
    ) -> Optional[str]:
        """Infer temporal relation between events."""
        # Check typical sequences
        for sequence in self._typical_sequences.values():
            idx1 = -1
            idx2 = -1

            for i, step in enumerate(sequence):
                if step in event1.lower() or event1.lower() in step:
                    idx1 = i
                if step in event2.lower() or event2.lower() in step:
                    idx2 = i

            if idx1 >= 0 and idx2 >= 0:
                if idx1 < idx2:
                    return "before"
                elif idx1 > idx2:
                    return "after"
                else:
                    return "simultaneous"

        return None

    def is_plausible_duration(
        self,
        activity: str,
        duration: float,
    ) -> bool:
        """Check if duration is plausible for activity."""
        typical = self.get_typical_duration(activity)
        if typical is None:
            return True  # Unknown activity, assume plausible

        min_dur, max_dur = typical
        # Allow some flexibility
        return min_dur * 0.5 <= duration <= max_dur * 2


class CommonSenseReasoner:
    """
    Unified common sense reasoning system.

    Combines physics, social, and temporal reasoning.
    """

    def __init__(self):
        self.kb = CommonSenseKB()
        self.physics = PhysicsReasoner(self.kb)
        self.social = SocialReasoner(self.kb)
        self.temporal = TemporalReasoner()

    def reason(
        self,
        query: str,
        context: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:
        """Perform common sense reasoning on a query."""
        result = {
            "query": query,
            "physics_effects": [],
            "social_norms": [],
            "temporal_info": None,
        }

        # Physics reasoning
        if context and "object" in context:
            effects = self.physics.predict_effect(context["object"], query)
            result["physics_effects"] = effects

        # Social reasoning
        norms = self.social.get_applicable_norms(query)
        result["social_norms"] = [{"name": n.name, "expected": n.expected_behavior} for n in norms]

        # Temporal reasoning
        if context and "activity" in context:
            duration = self.temporal.get_typical_duration(context["activity"])
            if duration:
                result["temporal_info"] = {
                    "typical_duration_min": duration[0],
                    "typical_duration_max": duration[1],
                }

        return result

    def is_plausible(
        self,
        statement: str,
        context: Optional[Dict[str, Any]] = None,
    ) -> Tuple[bool, float, str]:
        """Check if a statement is plausible."""
        # Simple plausibility checks
        statement_lower = statement.lower()

        # Physics violations
        if "fly" in statement_lower and "human" in statement_lower:
            if "airplane" not in statement_lower:
                return False, 0.1, "Humans cannot fly without equipment"

        if "underwater" in statement_lower and "breathe" in statement_lower:
            if "fish" not in statement_lower and "scuba" not in statement_lower:
                return False, 0.1, "Most creatures cannot breathe underwater"

        # Temporal violations
        if "before born" in statement_lower or "before birth" in statement_lower:
            if "remember" in statement_lower:
                return False, 0.05, "Cannot remember events before birth"

        return True, 0.7, "No obvious violations detected"

    def fill_gaps(
        self,
        partial_story: List[str],
    ) -> List[str]:
        """Fill in missing common sense details in a story."""
        filled = []

        for i, event in enumerate(partial_story):
            filled.append(event)

            # Check if we need to add implicit events
            if i < len(partial_story) - 1:
                next_event = partial_story[i + 1]

                # Travel inference
                if "at home" in event.lower() and "at work" in next_event.lower():
                    filled.append("(implicit: traveled to work)")

                # Preparation inference
                if "eat" in next_event.lower() and "cook" not in event.lower():
                    filled.append("(implicit: prepared or obtained food)")

        return filled

    def statistics(self) -> Dict[str, Any]:
        """Get reasoner statistics."""
        return {
            "physics_rules": len(self.physics._rules),
            "social_norms": len(self.social._norms),
            "typical_durations": len(self.temporal._typical_durations),
            "typical_sequences": len(self.temporal._typical_sequences),
        }



"""
NEURO Streaming Engine - Real-time LLM Response Streaming

Provides async streaming of tokens from Ollama for instant feedback.
Uses aiohttp for efficient async HTTP.
"""

import asyncio
import json
from typing import AsyncGenerator, Optional, List, Dict, Any, Callable
from dataclasses import dataclass

try:
    import aiohttp

    AIOHTTP_AVAILABLE = True
except ImportError:
    AIOHTTP_AVAILABLE = False


@dataclass
class StreamConfig:
    """Configuration for streaming."""

    base_url: str = "http://localhost:11434"
    model: str = "ministral-3:8b"
    timeout: int = 120
    temperature: float = 0.7
    max_tokens: Optional[int] = None


@dataclass
class StreamChunk:
    """A single chunk from the stream."""

    content: str
    done: bool = False
    model: str = ""
    total_duration: Optional[int] = None
    eval_count: Optional[int] = None


class StreamHandler:
    """
    Handles streaming responses from Ollama.

    Features:
    - Real-time token streaming
    - Progress callbacks
    - Graceful error handling
    - Sync fallback if aiohttp unavailable
    """

    def __init__(self, config: Optional[StreamConfig] = None):
        self.config = config or StreamConfig()
        self._session: Optional[aiohttp.ClientSession] = None

    async def _get_session(self) -> aiohttp.ClientSession:
        """Get or create aiohttp session."""
        if self._session is None or self._session.closed:
            timeout = aiohttp.ClientTimeout(total=self.config.timeout)
            self._session = aiohttp.ClientSession(timeout=timeout)
        return self._session

    async def close(self):
        """Close the session."""
        if self._session and not self._session.closed:
            await self._session.close()

    async def stream(
        self,
        messages: List[Dict[str, str]],
        on_token: Optional[Callable[[str], None]] = None,
        on_done: Optional[Callable[[Dict], None]] = None,
    ) -> AsyncGenerator[StreamChunk, None]:
        """
        Stream tokens from Ollama.

        Args:
            messages: Chat messages in OpenAI format
            on_token: Callback for each token (for side effects like printing)
            on_done: Callback when stream completes

        Yields:
            StreamChunk objects with content
        """
        if not AIOHTTP_AVAILABLE:
            # Fallback to sync if aiohttp not installed
            async for chunk in self._sync_fallback(messages):
                if on_token and chunk.content:
                    on_token(chunk.content)
                yield chunk
            return

        session = await self._get_session()
        url = f"{self.config.base_url}/api/chat"

        payload = {
            "model": self.config.model,
            "messages": messages,
            "stream": True,
            "options": {
                "temperature": self.config.temperature,
            },
        }

        if self.config.max_tokens:
            payload["options"]["num_predict"] = self.config.max_tokens

        try:
            async with session.post(url, json=payload) as response:
                if response.status != 200:
                    error_text = await response.text()
                    yield StreamChunk(content=f"Error: {response.status} - {error_text}", done=True)
                    return

                full_response = ""
                async for line in response.content:
                    if not line:
                        continue

                    try:
                        data = json.loads(line.decode("utf-8"))
                    except json.JSONDecodeError:
                        continue

                    # Extract content
                    content = data.get("message", {}).get("content", "")
                    done = data.get("done", False)

                    if content:
                        full_response += content
                        if on_token:
                            on_token(content)

                    chunk = StreamChunk(
                        content=content,
                        done=done,
                        model=data.get("model", ""),
                        total_duration=data.get("total_duration"),
                        eval_count=data.get("eval_count"),
                    )

                    yield chunk

                    if done:
                        if on_done:
                            on_done(
                                {
                                    "full_response": full_response,
                                    "model": chunk.model,
                                    "total_duration": chunk.total_duration,
                                    "eval_count": chunk.eval_count,
                                }
                            )
                        break

        except aiohttp.ClientError as e:
            yield StreamChunk(content=f"Connection error: {e}", done=True)
        except asyncio.TimeoutError:
            yield StreamChunk(content="Request timed out", done=True)

    async def _sync_fallback(
        self, messages: List[Dict[str, str]]
    ) -> AsyncGenerator[StreamChunk, None]:
        """Sync fallback using requests (yields chunks from full response)."""
        import requests

        url = f"{self.config.base_url}/api/chat"
        payload = {
            "model": self.config.model,
            "messages": messages,
            "stream": False,
        }

        try:
            response = requests.post(url, json=payload, timeout=self.config.timeout)
            if response.status_code == 200:
                data = response.json()
                content = data.get("message", {}).get("content", "")
                # Yield content in chunks to simulate streaming
                words = content.split()
                for i, word in enumerate(words):
                    yield StreamChunk(
                        content=word + (" " if i < len(words) - 1 else ""),
                        done=(i == len(words) - 1),
                        model=data.get("model", ""),
                    )
            else:
                yield StreamChunk(content=f"Error: {response.status_code}", done=True)
        except Exception as e:
            yield StreamChunk(content=f"Error: {e}", done=True)

    async def stream_with_context(
        self,
        query: str,
        system_prompt: str,
        context: Optional[str] = None,
        history: Optional[List[Dict[str, str]]] = None,
    ) -> AsyncGenerator[StreamChunk, None]:
        """
        Stream with full context building.

        Args:
            query: User's query
            system_prompt: System instructions
            context: Additional context (from cognitive pipeline)
            history: Conversation history
        """
        messages = []

        # Build system message
        full_system = system_prompt
        if context:
            full_system += f"\n\n[Context]\n{context}"

        messages.append({"role": "system", "content": full_system})

        # Add history
        if history:
            messages.extend(history)

        # Add current query
        messages.append({"role": "user", "content": query})

        async for chunk in self.stream(messages):
            yield chunk


class TerminalStreamer:
    """
    Pretty terminal output for streaming.

    Handles:
    - Real-time character printing
    - Progress indicators
    - Color support
    """

    CYAN = "\033[36m"
    GREEN = "\033[32m"
    YELLOW = "\033[33m"
    DIM = "\033[2m"
    RESET = "\033[0m"
    BOLD = "\033[1m"

    def __init__(self, show_thinking: bool = True):
        self.show_thinking = show_thinking
        self.token_count = 0
        self.start_time = None

    def on_start(self, phase: str = "Thinking"):
        """Called when streaming starts."""
        import time

        self.start_time = time.time()
        self.token_count = 0
        if self.show_thinking:
            print(f"\n  {self.CYAN}{self.BOLD}Neuro:{self.RESET} ", end="", flush=True)

    def on_token(self, token: str):
        """Called for each token."""
        self.token_count += 1
        print(token, end="", flush=True)

    def on_done(self, stats: Dict[str, Any]):
        """Called when streaming completes."""
        import time

        duration = time.time() - self.start_time if self.start_time else 0
        print()  # New line after response
        if self.show_thinking:
            tokens_per_sec = self.token_count / max(0.1, duration)
            print(
                f"  {self.DIM}[{self.token_count} tokens, {tokens_per_sec:.1f} tok/s]{self.RESET}"
            )

    async def stream_to_terminal(
        self,
        handler: StreamHandler,
        messages: List[Dict[str, str]],
    ) -> str:
        """Stream response directly to terminal."""
        self.on_start()
        full_response = ""

        async for chunk in handler.stream(messages, on_token=self.on_token):
            full_response += chunk.content
            if chunk.done:
                self.on_done({"eval_count": chunk.eval_count})

        return full_response


# Convenience functions for simple usage


async def stream_chat(
    query: str,
    system_prompt: str = "You are a helpful AI assistant.",
    model: str = "ministral-3:8b",
    print_tokens: bool = True,
) -> str:
    """
    Simple streaming chat function.

    Args:
        query: User's query
        system_prompt: System instructions
        model: Ollama model to use
        print_tokens: Whether to print tokens as they arrive

    Returns:
        Full response text
    """
    config = StreamConfig(model=model)
    handler = StreamHandler(config)

    messages = [
        {"role": "system", "content": system_prompt},
        {"role": "user", "content": query},
    ]

    full_response = ""

    def on_token(token: str):
        nonlocal full_response
        full_response += token
        if print_tokens:
            print(token, end="", flush=True)

    async for chunk in handler.stream(messages, on_token=on_token):
        if chunk.done:
            break

    await handler.close()

    if print_tokens:
        print()  # New line

    return full_response


def stream_chat_sync(
    query: str,
    system_prompt: str = "You are a helpful AI assistant.",
    model: str = "ministral-3:8b",
    print_tokens: bool = True,
) -> str:
    """Synchronous wrapper for stream_chat."""
    return asyncio.run(stream_chat(query, system_prompt, model, print_tokens))


# =============================================================================
# TEST
# =============================================================================

if __name__ == "__main__":
    print("=" * 60)
    print("NEURO STREAMING ENGINE TEST")
    print("=" * 60)

    async def test():
        # Test basic streaming
        print("\nTest 1: Basic streaming")
        print("-" * 40)

        response = await stream_chat("What is 2 + 2? Answer in one sentence.", print_tokens=True)

        print(f"\nFull response: {response}")

        # Test with context
        print("\n" + "=" * 60)
        print("Test 2: Streaming with cognitive context")
        print("-" * 40)

        config = StreamConfig()
        handler = StreamHandler(config)
        streamer = TerminalStreamer(show_thinking=True)

        messages = [
            {
                "role": "system",
                "content": "You are NEURO, a neuroscience-inspired AI. Be helpful and direct.",
            },
            {"role": "user", "content": "Explain recursion briefly."},
        ]

        full = await streamer.stream_to_terminal(handler, messages)
        await handler.close()

        print(f"\n\nTotal length: {len(full)} characters")

    asyncio.run(test())



"""
Reward-Modulated Learning

Dopamine-modulated STDP: weight changes only when rewarded
"""

import numpy as np
from typing import Optional, List
from dataclasses import dataclass

from .stdp import STDPParams, STDPNetwork


@dataclass
class DopamineParams:
    """Dopamine system parameters"""

    baseline: float = 0.5  # Baseline dopamine level
    release_rate: float = 1.0  # Release per reward unit
    decay_rate: float = 0.1  # Decay rate
    prediction_rate: float = 0.1  # Reward prediction learning rate


class DopamineSystem:
    """Simplified dopamine system for reward signaling

    Dopamine signals reward prediction error:
    δ = reward - expected_reward
    """

    def __init__(self, params: Optional[DopamineParams] = None):
        self.params = params or DopamineParams()
        self.level = self.params.baseline
        self.expected_reward = 0.0
        self.prediction_error = 0.0

        # History
        self.level_history: List[float] = []
        self.error_history: List[float] = []

    def receive_reward(self, reward: float) -> float:
        """Process reward and compute prediction error

        Args:
            reward: Actual reward received

        Returns:
            Dopamine level (proportional to prediction error)
        """
        # Prediction error
        self.prediction_error = reward - self.expected_reward

        # Update expected reward
        self.expected_reward += self.params.prediction_rate * self.prediction_error

        # Dopamine level reflects prediction error
        self.level = self.params.baseline + self.params.release_rate * self.prediction_error
        self.level = max(0.0, self.level)  # Can't go negative

        # Track history
        self.level_history.append(self.level)
        self.error_history.append(self.prediction_error)

        return self.level

    def update(self, dt: float = 1.0) -> float:
        """Decay dopamine toward baseline"""
        self.level = self.level + self.params.decay_rate * (self.params.baseline - self.level) * dt
        return self.level

    def get_modulation(self) -> float:
        """Get current modulation signal for learning"""
        # Positive error = enhance learning
        # Negative error = suppress learning
        return max(0.0, self.prediction_error)

    def reset(self) -> None:
        """Reset to baseline"""
        self.level = self.params.baseline
        self.expected_reward = 0.0
        self.prediction_error = 0.0


class RewardModulatedSTDP(STDPNetwork):
    """STDP with reward modulation (three-factor rule)

    Δw = dopamine × STDP_trace

    Weight changes only consolidate when reward arrives.
    """

    def __init__(
        self,
        n_pre: int,
        n_post: int,
        stdp_params: Optional[STDPParams] = None,
        dopamine_params: Optional[DopamineParams] = None,
    ):
        super().__init__(n_pre, n_post, stdp_params)

        self.dopamine = DopamineSystem(dopamine_params)

        # Eligibility trace (potential weight change waiting for reward)
        self.eligibility_trace = np.zeros((n_post, n_pre))
        self.eligibility_decay = 0.95

    def update(self, pre_spikes: np.ndarray, post_spikes: np.ndarray, dt: float = 1.0) -> None:
        """Update eligibility traces (not weights yet)"""
        # Update pre trace first (LTP uses updated pre_trace)
        self.pre_trace = self.pre_trace * self.trace_decay + pre_spikes

        # Compute STDP-based eligibility using old post_trace for LTD
        ltp = self.params.A_plus * np.outer(post_spikes, self.pre_trace)
        ltd = self.params.A_minus * np.outer(self.post_trace, pre_spikes)

        # Now update post trace
        self.post_trace = self.post_trace * self.trace_decay + post_spikes

        # Accumulate in eligibility trace
        self.eligibility_trace = self.eligibility_trace * self.eligibility_decay + (ltp - ltd)

        # Decay dopamine
        self.dopamine.update(dt)

    def receive_reward(self, reward: float) -> None:
        """Apply reward to consolidate eligibility traces into weights

        Args:
            reward: Reward signal (positive = good, negative = bad)
        """
        # Get dopamine modulation
        self.dopamine.receive_reward(reward)
        modulation = self.dopamine.get_modulation()

        # Apply modulated eligibility to weights
        self.weights += modulation * self.eligibility_trace
        self.weights = np.clip(self.weights, self.params.w_min, self.params.w_max)

        # Decay eligibility after reward
        self.eligibility_trace *= 0.5

    def train_episode(
        self, pre_sequence: np.ndarray, post_sequence: np.ndarray, reward: float
    ) -> float:
        """Train on sequence with delayed reward

        Args:
            pre_sequence: (T, n_pre) binary spikes
            post_sequence: (T, n_post) binary spikes
            reward: Reward at end of episode

        Returns:
            Total weight change
        """
        w_before = self.weights.copy()

        # Accumulate eligibility over sequence
        for pre, post in zip(pre_sequence, post_sequence):
            self.update(pre, post)

        # Apply reward at end
        self.receive_reward(reward)

        return np.mean(np.abs(self.weights - w_before))

    def reset(self) -> None:
        """Reset all traces"""
        self.reset_traces()
        self.eligibility_trace = np.zeros((self.n_post, self.n_pre))
        self.dopamine.reset()


class ErrorBasedLearning:
    """Error-based learning (supervised-like)

    Δw = η × error × input
    """

    def __init__(self, n_input: int, n_output: int, learning_rate: float = 0.01):
        self.n_input = n_input
        self.n_output = n_output
        self.learning_rate = learning_rate

        self.weights = np.random.randn(n_output, n_input) * 0.1

    def forward(self, x: np.ndarray) -> np.ndarray:
        """Compute output"""
        return np.tanh(self.weights @ x)

    def update(self, x: np.ndarray, target: np.ndarray) -> float:
        """Update weights based on error

        Args:
            x: Input
            target: Target output

        Returns:
            Error magnitude
        """
        output = self.forward(x)
        error = target - output

        # Gradient descent on error
        dW = self.learning_rate * np.outer(error, x)
        self.weights += dW

        return np.mean(error**2)

    def train(self, inputs: np.ndarray, targets: np.ndarray, n_epochs: int = 100) -> List[float]:
        """Train on dataset"""
        errors = []

        for epoch in range(n_epochs):
            epoch_error = 0.0
            for x, t in zip(inputs, targets):
                epoch_error += self.update(x, t)
            errors.append(epoch_error / len(inputs))

        return errors



"""
Ollama Client - Local LLM API Wrapper

Provides direct integration with Ollama for local model inference.
"""

import requests
import json
from dataclasses import dataclass
from typing import Optional, List, Dict, Iterator


@dataclass
class OllamaConfig:
    """Configuration for Ollama client."""

    base_url: str = "http://localhost:11434"
    default_model: str = "ministral-3:8b"  # Ministral 3 8B - ONLY model
    timeout: int = 120
    temperature: float = 0.7
    max_tokens: int = 2048


class OllamaClient:
    """
    Local LLM client using Ollama.

    Provides both streaming and non-streaming interfaces
    for local model inference.
    """

    def __init__(self, config: Optional[OllamaConfig] = None):
        self.config = config or OllamaConfig()

    def is_available(self) -> bool:
        """Check if Ollama is running and accessible."""
        try:
            response = requests.get(f"{self.config.base_url}/api/tags", timeout=5)
            return response.status_code == 200
        except Exception:
            return False

    def list_models(self) -> List[str]:
        """List available models."""
        try:
            response = requests.get(f"{self.config.base_url}/api/tags", timeout=10)
            if response.status_code == 200:
                data = response.json()
                return [m["name"] for m in data.get("models", [])]
            return []
        except Exception:
            return []

    def generate(
        self,
        prompt: str,
        model: Optional[str] = None,
        system: Optional[str] = None,
        temperature: Optional[float] = None,
        max_tokens: Optional[int] = None,
    ) -> str:
        """
        Generate response from local model.

        Args:
            prompt: The user prompt
            model: Model to use (defaults to config default)
            system: Optional system prompt
            temperature: Sampling temperature
            max_tokens: Maximum tokens to generate

        Returns:
            Generated text response
        """
        model = model or self.config.default_model
        temperature = temperature if temperature is not None else self.config.temperature

        payload = {
            "model": model,
            "prompt": prompt,
            "stream": False,
            "options": {
                "temperature": temperature,
            },
        }

        if max_tokens:
            payload["options"]["num_predict"] = max_tokens
        if system:
            payload["system"] = system

        response = requests.post(
            f"{self.config.base_url}/api/generate", json=payload, timeout=self.config.timeout
        )

        if response.status_code != 200:
            raise RuntimeError(f"Ollama error: {response.text}")

        return response.json()["response"]

    def generate_stream(
        self,
        prompt: str,
        model: Optional[str] = None,
        system: Optional[str] = None,
        temperature: Optional[float] = None,
    ) -> Iterator[str]:
        """
        Generate response with streaming.

        Yields:
            Tokens as they are generated
        """
        model = model or self.config.default_model
        temperature = temperature if temperature is not None else self.config.temperature

        payload = {
            "model": model,
            "prompt": prompt,
            "stream": True,
            "options": {
                "temperature": temperature,
            },
        }

        if system:
            payload["system"] = system

        response = requests.post(
            f"{self.config.base_url}/api/generate",
            json=payload,
            timeout=self.config.timeout,
            stream=True,
        )

        if response.status_code != 200:
            raise RuntimeError(f"Ollama error: {response.text}")

        for line in response.iter_lines():
            if line:
                data = json.loads(line)
                if "response" in data:
                    yield data["response"]
                if data.get("done", False):
                    break

    def chat(
        self,
        messages: List[Dict[str, str]],
        model: Optional[str] = None,
        temperature: Optional[float] = None,
    ) -> str:
        """
        Chat completion with local model.

        Args:
            messages: List of {"role": "user/assistant/system", "content": "..."}
            model: Model to use
            temperature: Sampling temperature

        Returns:
            Assistant response
        """
        model = model or self.config.default_model
        temperature = temperature if temperature is not None else self.config.temperature

        response = requests.post(
            f"{self.config.base_url}/api/chat",
            json={
                "model": model,
                "messages": messages,
                "stream": False,
                "options": {
                    "temperature": temperature,
                },
            },
            timeout=self.config.timeout,
        )

        if response.status_code != 200:
            raise RuntimeError(f"Ollama error: {response.text}")

        return response.json()["message"]["content"]

    def chat_stream(
        self,
        messages: List[Dict[str, str]],
        model: Optional[str] = None,
        temperature: Optional[float] = None,
    ) -> Iterator[str]:
        """
        Chat completion with streaming.

        Yields:
            Tokens as they are generated
        """
        model = model or self.config.default_model
        temperature = temperature if temperature is not None else self.config.temperature

        response = requests.post(
            f"{self.config.base_url}/api/chat",
            json={
                "model": model,
                "messages": messages,
                "stream": True,
                "options": {
                    "temperature": temperature,
                },
            },
            timeout=self.config.timeout,
            stream=True,
        )

        if response.status_code != 200:
            raise RuntimeError(f"Ollama error: {response.text}")

        for line in response.iter_lines():
            if line:
                data = json.loads(line)
                if "message" in data and "content" in data["message"]:
                    yield data["message"]["content"]
                if data.get("done", False):
                    break



"""
Motor Interface: Generates output actions from cognitive state.

Converts internal cognitive representations to:
- Vector outputs (continuous actions)
- Discrete outputs (categorical actions)
- Text outputs (language generation)
"""

from dataclasses import dataclass, field
from typing import Dict, List, Optional, Any
from enum import Enum
import numpy as np
import time

from .config import SystemConfig


class OutputType(Enum):
    """Types of motor output."""

    VECTOR = "vector"
    DISCRETE = "discrete"
    TEXT = "text"
    COMPOSITE = "composite"


class ActionCategory(Enum):
    """Categories of discrete actions."""

    MOVE = "move"
    INTERACT = "interact"
    COMMUNICATE = "communicate"
    WAIT = "wait"
    EXPLORE = "explore"


@dataclass
class MotorOutput:
    """A single motor output."""

    output_type: OutputType
    value: Any
    confidence: float = 0.5
    priority: float = 0.5
    timestamp: float = field(default_factory=time.time)
    metadata: Dict[str, Any] = field(default_factory=dict)


@dataclass
class ActionBuffer:
    """Buffer for recent actions."""

    capacity: int = 10
    actions: List[MotorOutput] = field(default_factory=list)

    def add(self, action: MotorOutput) -> None:
        """Add action to buffer."""
        self.actions.append(action)
        if len(self.actions) > self.capacity:
            self.actions.pop(0)

    def get_recent(self, n: int = 1) -> List[MotorOutput]:
        """Get n most recent actions."""
        return self.actions[-n:]

    def clear(self) -> None:
        """Clear buffer."""
        self.actions = []


class MotorInterface:
    """
    Generates motor outputs from cognitive state.

    The interface:
    1. Receives action proposals from modules
    2. Selects and prioritizes actions
    3. Converts to appropriate output format
    4. Maintains action history
    """

    def __init__(self, config: Optional[SystemConfig] = None):
        self.config = config or SystemConfig()
        self.buffer = ActionBuffer(capacity=10)

        # Pending actions from modules
        self._pending_actions: List[MotorOutput] = []

        # Statistics
        self._action_count = 0
        self._last_action: Optional[MotorOutput] = None

    def propose_action(
        self,
        action: np.ndarray,
        confidence: float = 0.5,
        priority: float = 0.5,
        output_type: OutputType = OutputType.VECTOR,
        metadata: Optional[Dict] = None,
    ) -> None:
        """
        Propose an action from a module.

        Args:
            action: Action vector or value
            confidence: Confidence in the action
            priority: Priority level
            output_type: Type of output
            metadata: Additional information
        """
        output = MotorOutput(
            output_type=output_type,
            value=action,
            confidence=confidence,
            priority=priority,
            metadata=metadata or {},
        )
        self._pending_actions.append(output)

    def generate(
        self,
        cognitive_state: np.ndarray,
        output_type: OutputType = OutputType.VECTOR,
    ) -> MotorOutput:
        """
        Generate motor output from cognitive state.

        Args:
            cognitive_state: Current cognitive state vector
            output_type: Desired output type

        Returns:
            Motor output
        """
        if output_type == OutputType.VECTOR:
            return self._generate_vector(cognitive_state)
        elif output_type == OutputType.DISCRETE:
            return self._generate_discrete(cognitive_state)
        elif output_type == OutputType.TEXT:
            return self._generate_text(cognitive_state)
        elif output_type == OutputType.COMPOSITE:
            return self._generate_composite(cognitive_state)
        else:
            return self._generate_vector(cognitive_state)

    def _generate_vector(self, state: np.ndarray) -> MotorOutput:
        """Generate continuous action vector."""
        # Transform state to action space
        action = self._transform_to_action(state)

        # Compute confidence from state activation
        confidence = float(np.tanh(np.mean(np.abs(state))))

        output = MotorOutput(
            output_type=OutputType.VECTOR,
            value=action,
            confidence=confidence,
            priority=confidence,
        )

        self._record_action(output)
        return output

    def _generate_discrete(self, state: np.ndarray) -> MotorOutput:
        """Generate discrete categorical action."""
        # Partition state into action categories
        n_categories = len(ActionCategory)
        partition_size = len(state) // n_categories

        category_scores = []
        for i in range(n_categories):
            start = i * partition_size
            end = start + partition_size
            score = float(np.mean(state[start:end]))
            category_scores.append(score)

        # Select highest scoring category
        best_idx = np.argmax(category_scores)
        best_category = list(ActionCategory)[best_idx]

        confidence = float(np.max(category_scores) - np.mean(category_scores))
        confidence = np.clip(confidence + 0.5, 0, 1)

        output = MotorOutput(
            output_type=OutputType.DISCRETE,
            value=best_category,
            confidence=confidence,
            priority=confidence,
            metadata={"scores": dict(zip([c.value for c in ActionCategory], category_scores))},
        )

        self._record_action(output)
        return output

    def _generate_text(self, state: np.ndarray) -> MotorOutput:
        """Generate text output from cognitive state."""
        # Simple text generation from state
        # In production, would use language model

        # Use state to select words/tokens
        text_parts = []

        # Map state activation to basic tokens
        tokens = [
            "observe",
            "move",
            "wait",
            "think",
            "act",
            "explore",
            "remember",
            "decide",
            "communicate",
            "learn",
        ]

        # Select tokens based on state
        n_tokens = min(5, max(1, int(np.sum(np.abs(state)) / 10)))
        indices = np.argsort(state)[-n_tokens:]

        for idx in indices:
            token_idx = idx % len(tokens)
            text_parts.append(tokens[token_idx])

        text = " ".join(text_parts)

        confidence = float(np.mean(np.abs(state[:10])))
        confidence = np.clip(confidence, 0, 1)

        output = MotorOutput(
            output_type=OutputType.TEXT,
            value=text,
            confidence=confidence,
            priority=confidence,
        )

        self._record_action(output)
        return output

    def _generate_composite(self, state: np.ndarray) -> MotorOutput:
        """Generate composite output with multiple components."""
        vector = self._transform_to_action(state)
        discrete = self._generate_discrete(state)
        text = self._generate_text(state)

        composite = {
            "vector": vector,
            "discrete": discrete.value,
            "text": text.value,
        }

        confidence = (discrete.confidence + text.confidence) / 2

        output = MotorOutput(
            output_type=OutputType.COMPOSITE,
            value=composite,
            confidence=confidence,
            priority=confidence,
        )

        self._record_action(output)
        return output

    def _transform_to_action(self, state: np.ndarray) -> np.ndarray:
        """Transform cognitive state to action vector."""
        # Resize to output dimension
        if len(state) >= self.config.output_dim:
            action = state[: self.config.output_dim]
        else:
            action = np.pad(state, (0, self.config.output_dim - len(state)))

        # Apply nonlinearity
        action = np.tanh(action)

        # Apply threshold
        action = np.where(np.abs(action) > self.config.action_threshold, action, 0.0)

        return action.astype(np.float32)

    def select_action(self) -> Optional[MotorOutput]:
        """
        Select best action from pending proposals.

        Uses priority and confidence to rank actions.
        """
        if not self._pending_actions:
            return None

        # Score actions
        scored = []
        for action in self._pending_actions:
            score = action.priority * action.confidence
            scored.append((score, action))

        # Sort by score
        scored.sort(key=lambda x: x[0], reverse=True)

        # Select best
        best = scored[0][1]

        # Clear pending
        self._pending_actions = []

        self._record_action(best)
        return best

    def _record_action(self, action: MotorOutput) -> None:
        """Record action in history."""
        self.buffer.add(action)
        self._last_action = action
        self._action_count += 1

    def get_last_action(self) -> Optional[MotorOutput]:
        """Get most recent action."""
        return self._last_action

    def get_action_history(self, n: int = 5) -> List[MotorOutput]:
        """Get recent action history."""
        return self.buffer.get_recent(n)

    def get_statistics(self) -> Dict[str, Any]:
        """Get interface statistics."""
        return {
            "action_count": self._action_count,
            "pending_count": len(self._pending_actions),
            "buffer_size": len(self.buffer.actions),
            "output_dim": self.config.output_dim,
            "action_threshold": self.config.action_threshold,
        }

    def reset(self) -> None:
        """Reset interface state."""
        self.buffer.clear()
        self._pending_actions = []
        self._last_action = None
        self._action_count = 0



"""
Structure Mapping: Analogical reasoning via structure mapping.

Implements Gentner's Structure Mapping Theory for analogy.
"""

from dataclasses import dataclass, field
from typing import Dict, List, Optional, Any, Tuple, Set
from enum import Enum


class RelationOrder(Enum):
    """Order of relations (first-order, higher-order)."""

    FIRST = 1  # Relations between objects (e.g., larger(sun, moon))
    SECOND = 2  # Relations between relations (e.g., cause(heat, expand))
    THIRD = 3  # Relations between second-order relations


@dataclass
class Predicate:
    """A predicate in a relational structure."""

    name: str
    arguments: List[str]  # Object or relation names
    order: RelationOrder = RelationOrder.FIRST
    properties: Dict[str, Any] = field(default_factory=dict)

    def __hash__(self):
        return hash((self.name, tuple(self.arguments)))

    def arity(self) -> int:
        return len(self.arguments)


@dataclass
class RelationalStructure:
    """
    A relational structure representing a domain.

    Contains objects and relations between them.
    """

    name: str
    objects: Set[str] = field(default_factory=set)
    predicates: List[Predicate] = field(default_factory=list)
    attributes: Dict[str, Dict[str, Any]] = field(default_factory=dict)

    def add_object(self, name: str, attrs: Optional[Dict[str, Any]] = None) -> None:
        """Add an object to the structure."""
        self.objects.add(name)
        if attrs:
            self.attributes[name] = attrs

    def add_predicate(self, predicate: Predicate) -> None:
        """Add a predicate/relation."""
        self.predicates.append(predicate)
        for arg in predicate.arguments:
            self.objects.add(arg)

    def add_relation(
        self,
        name: str,
        arguments: List[str],
        order: RelationOrder = RelationOrder.FIRST,
    ) -> Predicate:
        """Convenience method to add a relation."""
        pred = Predicate(name, arguments, order)
        self.add_predicate(pred)
        return pred

    def get_predicates_involving(self, obj: str) -> List[Predicate]:
        """Get all predicates involving an object."""
        return [p for p in self.predicates if obj in p.arguments]


@dataclass
class Mapping:
    """A mapping between two objects/relations."""

    source: str
    target: str
    confidence: float = 1.0
    is_object: bool = True  # vs. relation mapping


@dataclass
class StructuralAlignment:
    """
    An alignment between two relational structures.

    Contains object and relation mappings.
    """

    source: RelationalStructure
    target: RelationalStructure
    object_mappings: Dict[str, str] = field(default_factory=dict)  # source -> target
    relation_mappings: Dict[str, str] = field(default_factory=dict)
    score: float = 0.0
    matched_predicates: List[Tuple[Predicate, Predicate]] = field(default_factory=list)

    def get_target_for(self, source_item: str) -> Optional[str]:
        """Get target mapping for a source item."""
        return self.object_mappings.get(source_item, self.relation_mappings.get(source_item))


@dataclass
class Analogy:
    """
    A complete analogy between two domains.

    Includes the alignment and any inferences drawn.
    """

    source_domain: RelationalStructure
    target_domain: RelationalStructure
    alignment: StructuralAlignment
    inferences: List[Predicate] = field(default_factory=list)  # New predicates inferred for target
    systematicity: float = 0.0  # How systematic the analogy is
    similarity: float = 0.0

    def explain(self) -> str:
        """Generate explanation of the analogy."""
        lines = [f"Analogy: {self.source_domain.name} -> {self.target_domain.name}"]
        lines.append(f"Systematicity: {self.systematicity:.2f}")
        lines.append(f"Similarity: {self.similarity:.2f}")

        lines.append("\nObject mappings:")
        for src, tgt in self.alignment.object_mappings.items():
            lines.append(f"  {src} -> {tgt}")

        if self.inferences:
            lines.append("\nInferences:")
            for pred in self.inferences:
                args = ", ".join(pred.arguments)
                lines.append(f"  {pred.name}({args})")

        return "\n".join(lines)


class StructureMapper:
    """
    Structure Mapping Engine for analogical reasoning.

    Implements Gentner's Structure Mapping Theory:
    - Systematicity principle: Prefer mappings with connected structure
    - One-to-one constraint: Each element maps to at most one
    - Parallel connectivity: If predicates match, arguments should match

    Supports:
    - Finding structural alignments
    - Drawing analogical inferences
    - Evaluating analogy quality
    """

    def __init__(
        self,
        relation_weight: float = 1.0,
        attribute_weight: float = 0.3,
        systematicity_weight: float = 2.0,
    ):
        self.relation_weight = relation_weight
        self.attribute_weight = attribute_weight
        self.systematicity_weight = systematicity_weight

    def map(
        self,
        source: RelationalStructure,
        target: RelationalStructure,
    ) -> StructuralAlignment:
        """
        Find the best structural alignment between source and target.

        Uses a greedy algorithm inspired by SME (Structure Mapping Engine).
        """
        # Find matching predicates (same name and arity)
        match_candidates = self._find_matching_predicates(source, target)

        # Build initial object mappings from predicate matches
        object_mappings, relation_mappings, matched_preds = self._build_mappings(
            match_candidates, source, target
        )

        # Score the alignment
        score = self._score_alignment(
            object_mappings, relation_mappings, matched_preds, source, target
        )

        return StructuralAlignment(
            source=source,
            target=target,
            object_mappings=object_mappings,
            relation_mappings=relation_mappings,
            score=score,
            matched_predicates=matched_preds,
        )

    def _find_matching_predicates(
        self,
        source: RelationalStructure,
        target: RelationalStructure,
    ) -> List[Tuple[Predicate, Predicate]]:
        """Find predicates that could potentially match."""
        matches = []

        for s_pred in source.predicates:
            for t_pred in target.predicates:
                # Same predicate name and arity
                if s_pred.name == t_pred.name and s_pred.arity() == t_pred.arity():
                    matches.append((s_pred, t_pred))

        return matches

    def _build_mappings(
        self,
        candidates: List[Tuple[Predicate, Predicate]],
        source: RelationalStructure,
        target: RelationalStructure,
    ) -> Tuple[Dict[str, str], Dict[str, str], List[Tuple[Predicate, Predicate]]]:
        """Build consistent mappings from predicate matches."""
        object_mappings = {}
        relation_mappings = {}
        matched = []

        # Sort by predicate order (prefer higher-order)
        candidates.sort(key=lambda x: x[0].order.value, reverse=True)

        for s_pred, t_pred in candidates:
            # Check if mapping is consistent with existing
            consistent = True
            proposed_maps = {}

            for s_arg, t_arg in zip(s_pred.arguments, t_pred.arguments):
                if s_arg in object_mappings:
                    if object_mappings[s_arg] != t_arg:
                        consistent = False
                        break
                else:
                    # Check one-to-one constraint
                    if t_arg in object_mappings.values():
                        consistent = False
                        break
                    proposed_maps[s_arg] = t_arg

            if consistent:
                object_mappings.update(proposed_maps)
                relation_mappings[s_pred.name] = t_pred.name
                matched.append((s_pred, t_pred))

        return object_mappings, relation_mappings, matched

    def _score_alignment(
        self,
        object_mappings: Dict[str, str],
        relation_mappings: Dict[str, str],
        matched_preds: List[Tuple[Predicate, Predicate]],
        source: RelationalStructure,
        target: RelationalStructure,
    ) -> float:
        """Score an alignment based on systematicity and coverage."""
        score = 0.0

        # Base score from matched predicates
        for s_pred, t_pred in matched_preds:
            # Higher-order predicates worth more
            order_bonus = s_pred.order.value * self.systematicity_weight
            score += self.relation_weight + order_bonus

        # Attribute similarity bonus
        for s_obj, t_obj in object_mappings.items():
            s_attrs = source.attributes.get(s_obj, {})
            t_attrs = target.attributes.get(t_obj, {})

            shared = set(s_attrs.keys()) & set(t_attrs.keys())
            for attr in shared:
                if s_attrs[attr] == t_attrs[attr]:
                    score += self.attribute_weight

        return score

    def make_analogy(
        self,
        source: RelationalStructure,
        target: RelationalStructure,
    ) -> Analogy:
        """
        Create a complete analogy with inferences.

        Finds alignment and projects unmapped source predicates to target.
        """
        alignment = self.map(source, target)

        # Generate inferences
        inferences = self._project_inferences(alignment, source, target)

        # Compute systematicity
        systematicity = self._compute_systematicity(alignment, source)

        # Compute overall similarity
        similarity = self._compute_similarity(alignment, source, target)

        return Analogy(
            source_domain=source,
            target_domain=target,
            alignment=alignment,
            inferences=inferences,
            systematicity=systematicity,
            similarity=similarity,
        )

    def _project_inferences(
        self,
        alignment: StructuralAlignment,
        source: RelationalStructure,
        target: RelationalStructure,
    ) -> List[Predicate]:
        """Project unmapped source predicates to target as inferences."""
        inferences = []
        mapped_source_preds = {s for s, _ in alignment.matched_predicates}

        for s_pred in source.predicates:
            if s_pred in mapped_source_preds:
                continue

            # Try to project this predicate
            new_args = []
            can_project = True

            for arg in s_pred.arguments:
                if arg in alignment.object_mappings:
                    new_args.append(alignment.object_mappings[arg])
                else:
                    can_project = False
                    break

            if can_project:
                inferred = Predicate(
                    name=s_pred.name,
                    arguments=new_args,
                    order=s_pred.order,
                )
                inferences.append(inferred)

        return inferences

    def _compute_systematicity(
        self,
        alignment: StructuralAlignment,
        source: RelationalStructure,
    ) -> float:
        """
        Compute systematicity of the mapping.

        Higher-order relations and connected structure increase systematicity.
        """
        if not alignment.matched_predicates:
            return 0.0

        total_order = sum(p[0].order.value for p in alignment.matched_predicates)
        max_order = len(alignment.matched_predicates) * 3  # Max is THIRD order

        return total_order / max_order if max_order > 0 else 0.0

    def _compute_similarity(
        self,
        alignment: StructuralAlignment,
        source: RelationalStructure,
        target: RelationalStructure,
    ) -> float:
        """Compute overall similarity between domains."""
        # Predicate coverage
        matched_count = len(alignment.matched_predicates)
        total_count = max(len(source.predicates), len(target.predicates))

        if total_count == 0:
            return 0.0

        return matched_count / total_count

    def evaluate_inference(
        self,
        inference: Predicate,
        target: RelationalStructure,
    ) -> float:
        """
        Evaluate plausibility of an analogical inference.

        Higher scores for inferences consistent with target structure.
        """
        # Check if objects exist in target
        if not all(arg in target.objects for arg in inference.arguments):
            return 0.0

        # Check if similar predicates exist
        similar = [
            p
            for p in target.predicates
            if p.name == inference.name or p.arity() == inference.arity()
        ]

        base_score = 0.5
        if similar:
            base_score += 0.3

        # Penalize if contradicts existing knowledge
        for p in target.predicates:
            if p.name == inference.name and p.arguments == inference.arguments:
                return 1.0  # Already exists, perfect match
            if p.name == inference.name and p.arity() == inference.arity():
                # Same predicate, different args - might contradict
                shared_args = set(p.arguments) & set(inference.arguments)
                if shared_args:
                    base_score -= 0.1

        return min(1.0, max(0.0, base_score))



"""
Pattern Recognition - Neocortex Simulation

Implements parallel pattern matching against learned statistical regularities.
This is the core of System 1's ability to recognize familiar situations instantly.

Based on research showing neocortex performs massively parallel pattern matching
without explicit rule representation - patterns are implicit in connection weights.
"""

import numpy as np
from typing import Dict, List, Optional, Any
from dataclasses import dataclass, field


@dataclass
class Pattern:
    """A learned pattern template"""

    id: str
    template: np.ndarray
    exemplar_count: int = 1
    variance: Optional[np.ndarray] = None
    contexts: List[Any] = field(default_factory=list)


@dataclass
class PatternMatch:
    """Result of pattern matching"""

    pattern_id: str
    confidence: float
    template: np.ndarray


class PatternRecognition:
    """
    Parallel pattern matching against learned regularities.

    Simulates neocortex statistical pattern detection:
    - All patterns checked SIMULTANEOUSLY (parallel)
    - No explicit rules - patterns are implicit
    - Context-dependent activation
    - Learns statistical regularities from examples
    """

    def __init__(self, similarity_threshold: float = 0.7, max_patterns: int = 10000):
        self.patterns: Dict[str, Pattern] = {}
        self.similarity_threshold = similarity_threshold
        self.max_patterns = max_patterns
        self._pattern_matrix: Optional[np.ndarray] = None
        self._pattern_ids: List[str] = []
        self._matrix_dirty = True

    def learn_pattern(
        self, pattern_id: str, examples: List[np.ndarray], contexts: Optional[List[Any]] = None
    ) -> Pattern:
        """
        Extract statistical regularity from examples.

        The brain learns patterns by extracting what's common across examples,
        not by storing explicit rules.
        """
        if len(examples) == 0:
            raise ValueError("Need at least one example to learn pattern")

        examples_array = np.array(examples)
        template = np.mean(examples_array, axis=0)
        variance = np.var(examples_array, axis=0) if len(examples) > 1 else None

        pattern = Pattern(
            id=pattern_id,
            template=template,
            exemplar_count=len(examples),
            variance=variance,
            contexts=contexts or [],
        )

        self.patterns[pattern_id] = pattern
        self._matrix_dirty = True

        return pattern

    def update_pattern(self, pattern_id: str, new_example: np.ndarray) -> Pattern:
        """
        Incrementally update pattern with new example.
        Online learning - brain continuously updates patterns.
        """
        if pattern_id not in self.patterns:
            return self.learn_pattern(pattern_id, [new_example])

        pattern = self.patterns[pattern_id]
        n = pattern.exemplar_count

        # Running mean update
        pattern.template = (pattern.template * n + new_example) / (n + 1)
        pattern.exemplar_count = n + 1

        self._matrix_dirty = True
        return pattern

    def _build_pattern_matrix(self):
        """Build matrix for vectorized parallel matching"""
        if not self._matrix_dirty:
            return

        self._pattern_ids = list(self.patterns.keys())
        if self._pattern_ids:
            self._pattern_matrix = np.array(
                [self.patterns[pid].template for pid in self._pattern_ids]
            )
        else:
            self._pattern_matrix = None
        self._matrix_dirty = False

    def match(
        self, input_vector: np.ndarray, top_k: Optional[int] = None, context: Optional[Any] = None
    ) -> List[PatternMatch]:
        """
        Find all matching patterns - PARALLEL operation.

        This is the key System 1 capability: all patterns are checked
        simultaneously, not sequentially. Returns in order of confidence.
        """
        self._build_pattern_matrix()

        if self._pattern_matrix is None or len(self._pattern_ids) == 0:
            return []

        # Vectorized similarity computation - ALL patterns at once
        # This simulates parallel neural processing
        input_normalized = input_vector / (np.linalg.norm(input_vector) + 1e-8)
        patterns_normalized = self._pattern_matrix / (
            np.linalg.norm(self._pattern_matrix, axis=1, keepdims=True) + 1e-8
        )

        similarities = np.dot(patterns_normalized, input_normalized)

        # Context modulation - patterns associated with current context get boost
        if context is not None:
            for i, pid in enumerate(self._pattern_ids):
                if context in self.patterns[pid].contexts:
                    similarities[i] *= 1.2  # Context boost

        # Filter by threshold
        matches = []
        for i, (pid, sim) in enumerate(zip(self._pattern_ids, similarities)):
            if sim >= self.similarity_threshold:
                matches.append(
                    PatternMatch(
                        pattern_id=pid, confidence=float(sim), template=self.patterns[pid].template
                    )
                )

        # Sort by confidence (highest first)
        matches.sort(key=lambda m: m.confidence, reverse=True)

        if top_k is not None:
            matches = matches[:top_k]

        return matches

    def confidence(self, input_vector: np.ndarray, pattern_id: str) -> float:
        """Get confidence score for specific pattern match"""
        if pattern_id not in self.patterns:
            return 0.0

        pattern = self.patterns[pattern_id]
        input_norm = input_vector / (np.linalg.norm(input_vector) + 1e-8)
        pattern_norm = pattern.template / (np.linalg.norm(pattern.template) + 1e-8)

        return float(np.dot(input_norm, pattern_norm))

    def best_match(self, input_vector: np.ndarray) -> Optional[PatternMatch]:
        """Get single best matching pattern"""
        matches = self.match(input_vector, top_k=1)
        return matches[0] if matches else None

    def has_match(self, input_vector: np.ndarray) -> bool:
        """Quick check if any pattern matches"""
        return len(self.match(input_vector, top_k=1)) > 0

    def generalize(self, novel_input: np.ndarray) -> List[PatternMatch]:
        """
        Generalization - find partial matches for novel input.

        System 1 can recognize things it hasn't seen exactly before
        by matching to similar known patterns.
        """
        # Lower threshold for generalization
        original_threshold = self.similarity_threshold
        self.similarity_threshold = original_threshold * 0.7

        matches = self.match(novel_input)

        self.similarity_threshold = original_threshold
        return matches



"""
Curriculum: Order tasks by complexity for progressive learning.

Implements curriculum learning strategies for optimal knowledge transfer.
"""

from dataclasses import dataclass, field
from typing import Dict, List, Optional, Any, Set, Callable
from enum import Enum
import numpy as np


class TaskDifficulty(Enum):
    """Task difficulty levels."""

    TRIVIAL = 1
    EASY = 2
    MEDIUM = 3
    HARD = 4
    EXPERT = 5


@dataclass
class Task:
    """A learning task."""

    id: str
    name: str
    domain: str
    difficulty: TaskDifficulty
    prerequisites: List[str] = field(default_factory=list)
    skills_required: List[str] = field(default_factory=list)
    skills_taught: List[str] = field(default_factory=list)
    reward_signal: Optional[Callable[[], float]] = None
    metadata: Dict[str, Any] = field(default_factory=dict)


@dataclass
class LearningPath:
    """A sequence of tasks for learning."""

    id: str
    name: str
    tasks: List[str]  # Task IDs in order
    estimated_difficulty_curve: List[float]
    skills_progression: List[str]


@dataclass
class LearnerState:
    """Current state of a learner."""

    skills_mastered: Set[str] = field(default_factory=set)
    tasks_completed: List[str] = field(default_factory=list)
    current_performance: Dict[str, float] = field(default_factory=dict)
    learning_rate: float = 1.0
    frustration_level: float = 0.0


class ProgressTracker:
    """
    Track learning progress across tasks.
    """

    def __init__(self):
        self._history: List[Dict[str, Any]] = []
        self._performance_by_task: Dict[str, List[float]] = {}
        self._performance_by_skill: Dict[str, List[float]] = {}

    def record(
        self,
        task_id: str,
        performance: float,
        skills_used: List[str],
        time_taken: float,
    ) -> None:
        """Record a task attempt."""
        self._history.append(
            {
                "task_id": task_id,
                "performance": performance,
                "skills_used": skills_used,
                "time_taken": time_taken,
            }
        )

        if task_id not in self._performance_by_task:
            self._performance_by_task[task_id] = []
        self._performance_by_task[task_id].append(performance)

        for skill in skills_used:
            if skill not in self._performance_by_skill:
                self._performance_by_skill[skill] = []
            self._performance_by_skill[skill].append(performance)

    def get_skill_mastery(self, skill: str) -> float:
        """Get estimated mastery of a skill."""
        performances = self._performance_by_skill.get(skill, [])
        if not performances:
            return 0.0

        # Weighted average with recency bias
        weights = np.exp(np.linspace(-2, 0, len(performances)))
        weights = weights / weights.sum()

        return float(np.dot(performances, weights))

    def get_task_performance(self, task_id: str) -> float:
        """Get best performance on a task."""
        performances = self._performance_by_task.get(task_id, [])
        if not performances:
            return 0.0
        return max(performances)

    def get_learning_curve(self, skill: str) -> List[float]:
        """Get learning curve for a skill."""
        return self._performance_by_skill.get(skill, [])

    def is_skill_mastered(
        self,
        skill: str,
        threshold: float = 0.8,
    ) -> bool:
        """Check if a skill is mastered."""
        return self.get_skill_mastery(skill) >= threshold


class DifficultyEstimator:
    """
    Estimate task difficulty based on features.
    """

    def __init__(self):
        self._difficulty_weights = {
            "n_prerequisites": 0.2,
            "n_skills_required": 0.3,
            "complexity": 0.3,
            "novelty": 0.2,
        }

    def estimate(
        self,
        task: Task,
        learner_state: LearnerState,
    ) -> float:
        """
        Estimate task difficulty for a learner.

        Returns:
            Difficulty score 0-1 (relative to learner's current state)
        """
        scores = {}

        # Prerequisite difficulty
        unmet = sum(1 for p in task.prerequisites if p not in learner_state.tasks_completed)
        scores["n_prerequisites"] = unmet / max(len(task.prerequisites), 1)

        # Skill gap
        missing = sum(1 for s in task.skills_required if s not in learner_state.skills_mastered)
        scores["n_skills_required"] = missing / max(len(task.skills_required), 1)

        # Base complexity from task difficulty
        scores["complexity"] = (task.difficulty.value - 1) / 4

        # Novelty (based on how many new skills it teaches)
        new_skills = sum(1 for s in task.skills_taught if s not in learner_state.skills_mastered)
        scores["novelty"] = new_skills / max(len(task.skills_taught), 1)

        # Weighted combination
        difficulty = sum(self._difficulty_weights[k] * scores[k] for k in scores)

        return min(1.0, max(0.0, difficulty))


class TaskSelector:
    """
    Select optimal next task based on curriculum.
    """

    def __init__(
        self,
        target_difficulty: float = 0.6,
        difficulty_tolerance: float = 0.2,
    ):
        self.target_difficulty = target_difficulty
        self.difficulty_tolerance = difficulty_tolerance
        self.difficulty_estimator = DifficultyEstimator()

    def select(
        self,
        available_tasks: List[Task],
        learner_state: LearnerState,
    ) -> Optional[Task]:
        """
        Select the best next task for the learner.
        """
        candidates = []

        for task in available_tasks:
            # Check prerequisites
            if not all(p in learner_state.tasks_completed for p in task.prerequisites):
                continue

            # Estimate difficulty
            difficulty = self.difficulty_estimator.estimate(task, learner_state)

            # Check if in target range
            if abs(difficulty - self.target_difficulty) <= self.difficulty_tolerance:
                score = self._score_task(task, learner_state, difficulty)
                candidates.append((task, score))

        if not candidates:
            # Fall back to easiest available task
            for task in available_tasks:
                if all(p in learner_state.tasks_completed for p in task.prerequisites):
                    difficulty = self.difficulty_estimator.estimate(task, learner_state)
                    candidates.append((task, 1.0 / (difficulty + 0.1)))

        if not candidates:
            return None

        # Return highest scoring task
        candidates.sort(key=lambda x: -x[1])
        return candidates[0][0]

    def _score_task(
        self,
        task: Task,
        learner_state: LearnerState,
        difficulty: float,
    ) -> float:
        """Score a task for selection."""
        score = 0.0

        # Prefer tasks teaching new skills
        new_skills = sum(1 for s in task.skills_taught if s not in learner_state.skills_mastered)
        score += new_skills * 0.5

        # Prefer tasks building on existing skills
        known_skills = sum(1 for s in task.skills_required if s in learner_state.skills_mastered)
        score += known_skills * 0.3

        # Penalty for being too far from target difficulty
        score -= abs(difficulty - self.target_difficulty) * 0.2

        return score


class CurriculumLearner:
    """
    Complete curriculum learning system.

    Manages task ordering, difficulty progression, and learning paths.
    """

    def __init__(
        self,
        initial_difficulty: float = 0.3,
        difficulty_increase_rate: float = 0.1,
        mastery_threshold: float = 0.8,
    ):
        self.mastery_threshold = mastery_threshold

        self.task_selector = TaskSelector(target_difficulty=initial_difficulty)
        self.progress_tracker = ProgressTracker()
        self.difficulty_estimator = DifficultyEstimator()

        # Task registry
        self._tasks: Dict[str, Task] = {}
        self._learning_paths: Dict[str, LearningPath] = {}

        # Current learner state
        self._learner_state = LearnerState()

        # Curriculum parameters
        self._difficulty_increase_rate = difficulty_increase_rate

    def register_task(self, task: Task) -> None:
        """Register a task in the curriculum."""
        self._tasks[task.id] = task

    def register_tasks(self, tasks: List[Task]) -> None:
        """Register multiple tasks."""
        for task in tasks:
            self.register_task(task)

    def create_learning_path(
        self,
        path_id: str,
        path_name: str,
        target_skills: List[str],
    ) -> LearningPath:
        """
        Create a learning path targeting specific skills.
        """
        # Find tasks that teach target skills
        relevant_tasks = []
        for task in self._tasks.values():
            if any(s in target_skills for s in task.skills_taught):
                relevant_tasks.append(task)

        # Topological sort by prerequisites
        ordered_tasks = self._topological_sort(relevant_tasks)

        # Estimate difficulty curve
        difficulty_curve = []
        for task in ordered_tasks:
            diff = (task.difficulty.value - 1) / 4
            difficulty_curve.append(diff)

        # Collect skills progression
        skills_progression = []
        seen_skills = set()
        for task in ordered_tasks:
            for skill in task.skills_taught:
                if skill not in seen_skills:
                    skills_progression.append(skill)
                    seen_skills.add(skill)

        path = LearningPath(
            id=path_id,
            name=path_name,
            tasks=[t.id for t in ordered_tasks],
            estimated_difficulty_curve=difficulty_curve,
            skills_progression=skills_progression,
        )

        self._learning_paths[path_id] = path
        return path

    def _topological_sort(self, tasks: List[Task]) -> List[Task]:
        """Sort tasks by prerequisites."""
        task_dict = {t.id: t for t in tasks}
        visited = set()
        result = []

        def visit(task_id: str):
            if task_id in visited:
                return
            visited.add(task_id)

            task = task_dict.get(task_id)
            if task:
                for prereq in task.prerequisites:
                    if prereq in task_dict:
                        visit(prereq)
                result.append(task)

        for task in tasks:
            visit(task.id)

        return result

    def get_next_task(self) -> Optional[Task]:
        """Get the next task for the learner."""
        available = list(self._tasks.values())
        return self.task_selector.select(available, self._learner_state)

    def complete_task(
        self,
        task_id: str,
        performance: float,
        time_taken: float = 0.0,
    ) -> Dict[str, Any]:
        """
        Record task completion.

        Returns:
            Summary of what was learned
        """
        task = self._tasks.get(task_id)
        if not task:
            raise ValueError(f"Task {task_id} not found")

        # Record in progress tracker
        self.progress_tracker.record(task_id, performance, task.skills_required, time_taken)

        # Update learner state
        self._learner_state.tasks_completed.append(task_id)
        self._learner_state.current_performance[task_id] = performance

        # Check for skill mastery
        newly_mastered = []
        for skill in task.skills_taught:
            if self.progress_tracker.is_skill_mastered(skill, self.mastery_threshold):
                if skill not in self._learner_state.skills_mastered:
                    self._learner_state.skills_mastered.add(skill)
                    newly_mastered.append(skill)

        # Update difficulty target if doing well
        if performance >= self.mastery_threshold:
            self.task_selector.target_difficulty = min(
                1.0, self.task_selector.target_difficulty + self._difficulty_increase_rate
            )
        elif performance < 0.5:
            self.task_selector.target_difficulty = max(
                0.2, self.task_selector.target_difficulty - self._difficulty_increase_rate
            )

        return {
            "task_id": task_id,
            "performance": performance,
            "newly_mastered_skills": newly_mastered,
            "total_skills_mastered": len(self._learner_state.skills_mastered),
            "current_difficulty_target": self.task_selector.target_difficulty,
        }

    def get_learner_state(self) -> LearnerState:
        """Get current learner state."""
        return self._learner_state

    def get_skill_progress(self, skill: str) -> Dict[str, Any]:
        """Get progress for a specific skill."""
        return {
            "skill": skill,
            "mastery": self.progress_tracker.get_skill_mastery(skill),
            "mastered": skill in self._learner_state.skills_mastered,
            "learning_curve": self.progress_tracker.get_learning_curve(skill),
        }

    def get_recommended_path(
        self,
        target_skills: List[str],
    ) -> Optional[LearningPath]:
        """Get a recommended learning path for target skills."""
        # Check if a suitable path exists
        for path in self._learning_paths.values():
            if all(s in path.skills_progression for s in target_skills):
                return path

        # Create a new path
        path_id = f"path_{len(self._learning_paths)}"
        return self.create_learning_path(path_id, f"Path to {target_skills}", target_skills)

    def reset(self) -> None:
        """Reset learner state."""
        self._learner_state = LearnerState()
        self.task_selector.target_difficulty = 0.3

    def statistics(self) -> Dict[str, Any]:
        """Get curriculum statistics."""
        return {
            "n_tasks": len(self._tasks),
            "n_paths": len(self._learning_paths),
            "tasks_completed": len(self._learner_state.tasks_completed),
            "skills_mastered": len(self._learner_state.skills_mastered),
            "current_difficulty": self.task_selector.target_difficulty,
        }



"""
neuro-robust: Robustness and Reliability for AGI Systems.

Provides:
- Adversarial training and defense
- Uncertainty quantification (epistemic vs aleatoric)
- Out-of-distribution (OOD) detection
- Calibrated confidence estimation
- Robust inference under uncertainty
"""

from .adversarial import (
    AdversarialAttack,
    AdversarialDefense,
    AttackType,
    DefenseType,
)
from .uncertainty import (
    UncertaintyQuantifier,
    UncertaintyType,
    EnsembleUncertainty,
)
from .ood_detection import (
    OODDetector,
    OODMethod,
    OODResult,
)
from .calibration import (
    ConfidenceCalibrator,
    CalibrationMethod,
    CalibrationMetrics,
)
from .robust_inference import (
    RobustInference,
    RobustPrediction,
    SelectivePrediction,
)
from .integration import (
    SharedSpaceRobustness,
    RobustEmbedding,
    RobustnessMetadata,
    RobustnessLevel,
)

__all__ = [
    # Adversarial
    "AdversarialAttack",
    "AdversarialDefense",
    "AttackType",
    "DefenseType",
    # Uncertainty
    "UncertaintyQuantifier",
    "UncertaintyType",
    "EnsembleUncertainty",
    # OOD Detection
    "OODDetector",
    "OODMethod",
    "OODResult",
    # Calibration
    "ConfidenceCalibrator",
    "CalibrationMethod",
    "CalibrationMetrics",
    # Robust Inference
    "RobustInference",
    "RobustPrediction",
    "SelectivePrediction",
    # Integration
    "SharedSpaceRobustness",
    "RobustEmbedding",
    "RobustnessMetadata",
    "RobustnessLevel",
]



"""
Memory Controller: Orchestrates the multi-store pipeline

Coordinates:
- Sensory → Working Memory → Long-Term Memory flow
- Attention gating
- Consolidation cycles
- Unified recall interface
"""

import time
from typing import Optional, Any, List, Dict
import numpy as np

from .sensory_memory import IconicBuffer, EchoicBuffer
from .working_memory import WorkingMemory
from .long_term_memory import EpisodicMemory, SemanticMemory, ProceduralMemory
from .memory_processes import (
    MemoryEncoder,
    MemoryConsolidator,
    MemoryRetriever,
    MemoryReconsolidator,
    Forgetter,
)


class MemoryController:
    """
    Orchestrates the multi-store memory pipeline.

    Implements the flow:
    Input → Sensory Memory → (attention) → Working Memory → (consolidation) → Long-Term Memory
    """

    def __init__(
        self, wm_capacity: int = 7, iconic_decay: float = 0.250, echoic_decay: float = 3.5
    ):
        """
        Initialize memory controller with all subsystems.

        Args:
            wm_capacity: Working memory capacity (5-9)
            iconic_decay: Iconic buffer decay time in seconds
            echoic_decay: Echoic buffer decay time in seconds
        """
        # Sensory memory
        self.iconic = IconicBuffer(decay_time=iconic_decay)
        self.echoic = EchoicBuffer(decay_time=echoic_decay)

        # Working memory
        self.working = WorkingMemory(capacity=wm_capacity)

        # Long-term memory stores
        self.episodic = EpisodicMemory()
        self.semantic = SemanticMemory()
        self.procedural = ProceduralMemory()

        # Memory processes
        self.encoder = MemoryEncoder()
        self.consolidator = MemoryConsolidator()
        self.retriever = MemoryRetriever()
        self.reconsolidator = MemoryReconsolidator()
        self.forgetter = Forgetter()

        # Time tracking
        self._time_fn = time.time
        self._last_tick = self._time_fn()

    def process_visual(self, visual_input: np.ndarray) -> None:
        """
        Process visual input into sensory memory.

        Args:
            visual_input: Visual data
        """
        self.iconic.capture(visual_input)

    def process_auditory(self, audio_input: np.ndarray) -> None:
        """
        Process auditory input into sensory memory.

        Args:
            audio_input: Audio data
        """
        self.echoic.capture(audio_input)

    def attend(self, modality: str = "visual") -> Optional[Any]:
        """
        Transfer attended sensory item to working memory.

        Args:
            modality: "visual" or "auditory"

        Returns:
            The attended item, or None if nothing available
        """
        if modality == "visual":
            data = self.iconic.read()
            if data is not None:
                encoded = self.encoder.encode_visual(data)
                self.working.store(encoded)
                return encoded

        elif modality == "auditory":
            data = self.echoic.read_window()
            if data is not None:
                encoded = self.encoder.encode_auditory(data)
                self.working.store(encoded)
                return encoded

        return None

    def store_in_working_memory(self, item: Any) -> bool:
        """
        Directly store item in working memory.

        Args:
            item: Item to store

        Returns:
            True if stored
        """
        return self.working.store(item)

    def consolidate(self) -> int:
        """
        Transfer working memory contents to long-term storage.

        Simulates sleep consolidation cycle.

        Returns:
            Number of items consolidated
        """
        # Get well-rehearsed items from working memory
        items = self.working.get_rehearsed_items()

        # Queue for consolidation
        for item in items:
            self.consolidator.queue_for_consolidation(item)

        # Run replay cycle
        consolidated = self.consolidator.replay_cycle()

        return consolidated

    def encode_experience(
        self,
        experience: Any,
        context: Optional[Dict[str, Any]] = None,
        emotional_valence: float = 0.0,
    ) -> Any:
        """
        Encode a new episodic memory.

        Args:
            experience: What happened
            context: Contextual information
            emotional_valence: Emotional coloring (-1 to +1)

        Returns:
            The created episode
        """
        episode = self.episodic.encode(
            experience=experience, context=context, emotional_valence=emotional_valence
        )

        # Also store in working memory for immediate access
        self.working.store(episode)

        return episode

    def learn_concept(
        self,
        name: str,
        features: Optional[Dict[str, float]] = None,
        relations: Optional[List[tuple]] = None,
    ) -> Any:
        """
        Add a concept to semantic memory.

        Args:
            name: Concept name
            features: Attribute-value pairs
            relations: List of (relation_type, target, strength)

        Returns:
            The created concept
        """
        return self.semantic.create_concept(name, features, relations)

    def learn_skill(
        self, name: str, trigger_features: Dict[str, Any], action_names: List[str]
    ) -> Any:
        """
        Encode a new procedural memory.

        Args:
            name: Skill name
            trigger_features: Features that trigger the skill
            action_names: Sequence of action names

        Returns:
            The created procedure
        """
        return self.procedural.encode_simple(name, trigger_features, action_names)

    def recall(self, cue: Any, memory_types: Optional[List[str]] = None) -> Optional[Any]:
        """
        Unified recall interface.

        Searches memory stores in order: working → episodic → semantic

        Args:
            cue: Query cue
            memory_types: List of store types to search (default: all)

        Returns:
            First matching memory, or None
        """
        if memory_types is None:
            memory_types = ["working", "episodic", "semantic"]

        for mem_type in memory_types:
            if mem_type == "working":
                result = self.working.retrieve(cue)
                if result is not None:
                    return result

            elif mem_type == "episodic":
                results = self.episodic.retrieve_by_cue(cue)
                if results:
                    return results[0]

            elif mem_type == "semantic":
                if isinstance(cue, str):
                    result = self.semantic.retrieve(cue)
                    if result is not None:
                        return result

        return None

    def recall_context(self, context: Dict[str, Any]) -> List[Any]:
        """
        Context-dependent recall from episodic memory.

        Args:
            context: Context features

        Returns:
            List of matching episodes
        """
        return self.episodic.retrieve_by_context(context)

    def recall_emotion(self, valence_min: float, valence_max: float) -> List[Any]:
        """
        Mood-congruent recall from episodic memory.

        Args:
            valence_min: Minimum emotional valence
            valence_max: Maximum emotional valence

        Returns:
            List of matching episodes
        """
        return self.episodic.retrieve_by_emotion(valence_min, valence_max)

    def execute_skill(self, stimulus: Dict[str, Any]) -> Optional[List[Any]]:
        """
        Execute matching procedural memory.

        Args:
            stimulus: Current stimulus features

        Returns:
            List of action results, or None
        """
        return self.procedural.execute(stimulus)

    def spread_concepts(self, start: str, depth: int = 3) -> Dict[str, float]:
        """
        Activate related concepts via spreading activation.

        Args:
            start: Starting concept
            depth: Propagation depth

        Returns:
            Dict of activated concepts with activation levels
        """
        return self.semantic.spread_activation(start, depth=depth)

    def update_memory(self, memory: Any, modification: Dict[str, Any]) -> Any:
        """
        Reconsolidate a memory with modifications.

        Args:
            memory: Memory to update
            modification: Changes to apply

        Returns:
            Updated memory
        """
        # Reactivate (makes labile)
        self.reconsolidator.reactivate(memory)

        # Modify
        updated = self.reconsolidator.modify(memory, modification)

        # Restabilize
        self.reconsolidator.restabilize(updated)

        return updated

    def forget(self, memory: Any, method: str = "decay") -> bool:
        """
        Apply forgetting to a memory.

        Args:
            memory: Memory to forget
            method: "decay", "suppress", or "prune"

        Returns:
            True if forgetting applied
        """
        if method == "decay":
            self.forgetter.decay(memory)
            return True
        elif method == "suppress":
            return self.forgetter.active_forget(memory)
        elif method == "prune" and hasattr(memory, "engram"):
            self.forgetter.prune(memory.engram)
            return True
        return False

    def tick(self, dt: Optional[float] = None) -> None:
        """
        Advance time and apply decay to all memory systems.

        Args:
            dt: Time step in seconds (auto-computed if None)
        """
        current_time = self._time_fn()

        if dt is None:
            dt = current_time - self._last_tick

        self._last_tick = current_time

        # Decay sensory buffers
        self.iconic.decay(current_time)
        self.echoic.decay(current_time)

        # Decay working memory
        self.working.decay_step(dt)

        # Decay procedural memory (skills weaken without practice)
        self.procedural.decay_all(dt * 0.001)  # Slow decay

    def get_state(self) -> Dict[str, Any]:
        """
        Get current state of all memory systems.

        Returns:
            Dict with state information
        """
        return {
            "iconic_available": self.iconic.is_available(),
            "echoic_duration": self.echoic.get_duration(),
            "working_memory_load": self.working.get_load(),
            "working_memory_items": len(self.working),
            "episodic_count": len(self.episodic),
            "semantic_count": len(self.semantic),
            "procedural_count": len(self.procedural),
        }

    def clear_working_memory(self) -> None:
        """Clear working memory"""
        self.working.clear()

    def clear_sensory_buffers(self) -> None:
        """Clear sensory memory buffers"""
        self.iconic.clear()
        self.echoic.clear()

    def set_time_function(self, time_fn) -> None:
        """Set custom time function for all components"""
        self._time_fn = time_fn
        self._last_tick = time_fn()

        self.iconic.set_time_function(time_fn)
        self.echoic.set_time_function(time_fn)
        self.working.set_time_function(time_fn)
        self.episodic.set_time_function(time_fn)
        self.semantic.set_time_function(time_fn)
        self.procedural.set_time_function(time_fn)
        self.encoder.set_time_function(time_fn)
        self.consolidator.set_time_function(time_fn)
        self.reconsolidator.set_time_function(time_fn)
        self.forgetter.set_time_function(time_fn)



"""
Skill Library

Manages reusable skills (parameterized options) for transfer and composition.
Skills can be learned, stored, retrieved, composed, and transferred across domains.
"""

from dataclasses import dataclass, field
from typing import Dict, List, Optional, Callable, Any, Tuple, Set
from enum import Enum
import numpy as np


class SkillType(Enum):
    """Types of skills."""

    PRIMITIVE = "primitive"
    COMPOSITE = "composite"
    PARAMETERIZED = "parameterized"
    ABSTRACT = "abstract"


@dataclass
class SkillMetadata:
    """Metadata about a skill."""

    created_at: float = 0.0
    last_used: float = 0.0
    use_count: int = 0
    success_count: int = 0
    average_steps: float = 0.0
    average_reward: float = 0.0
    domains: Set[str] = field(default_factory=set)
    tags: Set[str] = field(default_factory=set)

    @property
    def success_rate(self) -> float:
        """Compute success rate."""
        if self.use_count == 0:
            return 0.0
        return self.success_count / self.use_count

    def record_use(
        self,
        success: bool,
        steps: int,
        reward: float,
        timestamp: float,
    ) -> None:
        """Record a skill usage."""
        self.use_count += 1
        if success:
            self.success_count += 1
        self.last_used = timestamp

        n = self.use_count
        self.average_steps = self.average_steps * (n - 1) / n + steps / n
        self.average_reward = self.average_reward * (n - 1) / n + reward / n


@dataclass
class Skill:
    """
    A reusable skill that can be executed, composed, and transferred.

    Skills are more abstract than options - they represent capabilities
    that can be instantiated with parameters and adapted to contexts.
    """

    name: str
    skill_type: SkillType
    preconditions: Callable[[Any], bool]
    effects: Callable[[Any], Any]
    policy: Callable[[Any, Optional[Dict[str, Any]]], Any]
    termination: Callable[[Any, Optional[Dict[str, Any]]], bool]

    parameters: Dict[str, Any] = field(default_factory=dict)
    embedding: Optional[np.ndarray] = None
    metadata: SkillMetadata = field(default_factory=SkillMetadata)
    parent_skills: List[str] = field(default_factory=list)
    child_skills: List[str] = field(default_factory=list)

    def can_execute(self, state: Any, params: Optional[Dict[str, Any]] = None) -> bool:
        """Check if skill can be executed in state."""
        return self.preconditions(state)

    def execute_step(
        self,
        state: Any,
        params: Optional[Dict[str, Any]] = None,
    ) -> Tuple[Any, bool]:
        """
        Execute one step of the skill.

        Returns (action, terminated).
        """
        merged_params = {**self.parameters}
        if params:
            merged_params.update(params)

        action = self.policy(state, merged_params)
        terminated = self.termination(state, merged_params)

        return action, terminated

    def predict_effects(self, state: Any) -> Any:
        """Predict the effects of executing the skill."""
        return self.effects(state)

    def set_parameter(self, name: str, value: Any) -> None:
        """Set a skill parameter."""
        self.parameters[name] = value

    def get_parameter(self, name: str, default: Any = None) -> Any:
        """Get a skill parameter."""
        return self.parameters.get(name, default)

    def similarity(self, other: "Skill") -> float:
        """Compute similarity to another skill based on embeddings."""
        if self.embedding is None or other.embedding is None:
            return 0.0

        norm1 = np.linalg.norm(self.embedding)
        norm2 = np.linalg.norm(other.embedding)

        if norm1 < 1e-8 or norm2 < 1e-8:
            return 0.0

        return float(np.dot(self.embedding, other.embedding) / (norm1 * norm2))


@dataclass
class SkillComposer:
    """Composes multiple skills into composite skills."""

    def __init__(self, random_seed: Optional[int] = None):
        self._rng = np.random.default_rng(random_seed)

    def sequence(self, skills: List[Skill], name: str) -> Skill:
        """
        Create a sequential composite skill.

        Executes skills in order, each to completion.
        """
        current_skill_idx = [0]
        skill_done = [False]

        def combined_preconditions(state: Any) -> bool:
            if not skills:
                return False
            return skills[0].can_execute(state)

        def combined_effects(state: Any) -> Any:
            for skill in skills:
                state = skill.effects(state)
            return state

        def combined_policy(state: Any, params: Optional[Dict] = None) -> Any:
            if current_skill_idx[0] >= len(skills):
                return None

            current = skills[current_skill_idx[0]]
            action, done = current.execute_step(state, params)

            if done:
                skill_done[0] = True

            return action

        def combined_termination(state: Any, params: Optional[Dict] = None) -> bool:
            if skill_done[0]:
                skill_done[0] = False
                current_skill_idx[0] += 1
                if current_skill_idx[0] >= len(skills):
                    current_skill_idx[0] = 0
                    return True
            return False

        composite = Skill(
            name=name,
            skill_type=SkillType.COMPOSITE,
            preconditions=combined_preconditions,
            effects=combined_effects,
            policy=combined_policy,
            termination=combined_termination,
            child_skills=[s.name for s in skills],
        )

        if all(s.embedding is not None for s in skills):
            embeddings = [s.embedding for s in skills]
            composite.embedding = np.mean(embeddings, axis=0)

        return composite

    def parallel(self, skills: List[Skill], name: str) -> Skill:
        """
        Create a parallel composite skill.

        Executes all skills simultaneously, terminates when all complete.
        """
        skill_states = {}

        def combined_preconditions(state: Any) -> bool:
            return all(s.can_execute(state) for s in skills)

        def combined_effects(state: Any) -> Any:
            for skill in skills:
                state = skill.effects(state)
            return state

        def combined_policy(state: Any, params: Optional[Dict] = None) -> List[Any]:
            actions = []
            for skill in skills:
                if skill.name not in skill_states or not skill_states[skill.name]:
                    action, done = skill.execute_step(state, params)
                    skill_states[skill.name] = done
                    actions.append(action)
            return actions

        def combined_termination(state: Any, params: Optional[Dict] = None) -> bool:
            if all(skill_states.get(s.name, False) for s in skills):
                skill_states.clear()
                return True
            return False

        composite = Skill(
            name=name,
            skill_type=SkillType.COMPOSITE,
            preconditions=combined_preconditions,
            effects=combined_effects,
            policy=combined_policy,
            termination=combined_termination,
            child_skills=[s.name for s in skills],
        )

        return composite

    def conditional(
        self,
        condition: Callable[[Any], bool],
        if_skill: Skill,
        else_skill: Skill,
        name: str,
    ) -> Skill:
        """
        Create a conditional skill that branches based on state.
        """
        active_skill = [None]

        def combined_preconditions(state: Any) -> bool:
            return if_skill.can_execute(state) or else_skill.can_execute(state)

        def combined_effects(state: Any) -> Any:
            if condition(state):
                return if_skill.effects(state)
            return else_skill.effects(state)

        def combined_policy(state: Any, params: Optional[Dict] = None) -> Any:
            if active_skill[0] is None:
                active_skill[0] = if_skill if condition(state) else else_skill

            action, _ = active_skill[0].execute_step(state, params)
            return action

        def combined_termination(state: Any, params: Optional[Dict] = None) -> bool:
            if active_skill[0] is None:
                return False

            _, done = active_skill[0].execute_step(state, params)
            if done:
                active_skill[0] = None
                return True
            return False

        composite = Skill(
            name=name,
            skill_type=SkillType.COMPOSITE,
            preconditions=combined_preconditions,
            effects=combined_effects,
            policy=combined_policy,
            termination=combined_termination,
            child_skills=[if_skill.name, else_skill.name],
        )

        return composite


class SkillLibrary:
    """
    Library for storing, retrieving, and managing skills.

    Supports skill registration, similarity-based retrieval,
    composition, and cross-domain transfer.
    """

    def __init__(
        self,
        embedding_dim: int = 64,
        random_seed: Optional[int] = None,
    ):
        self.embedding_dim = embedding_dim
        self._rng = np.random.default_rng(random_seed)

        self._skills: Dict[str, Skill] = {}
        self._domain_skills: Dict[str, Set[str]] = {}
        self._tag_skills: Dict[str, Set[str]] = {}
        self._composer = SkillComposer(random_seed)

        self._current_time = 0.0

    def register_skill(self, skill: Skill) -> str:
        """
        Register a skill in the library.

        Returns the skill name.
        """
        if skill.embedding is None:
            skill.embedding = self._rng.normal(0, 1, self.embedding_dim)
            skill.embedding = skill.embedding / np.linalg.norm(skill.embedding)

        skill.metadata.created_at = self._current_time
        self._skills[skill.name] = skill

        for domain in skill.metadata.domains:
            if domain not in self._domain_skills:
                self._domain_skills[domain] = set()
            self._domain_skills[domain].add(skill.name)

        for tag in skill.metadata.tags:
            if tag not in self._tag_skills:
                self._tag_skills[tag] = set()
            self._tag_skills[tag].add(skill.name)

        return skill.name

    def get_skill(self, name: str) -> Optional[Skill]:
        """Get a skill by name."""
        return self._skills.get(name)

    def retrieve_applicable(
        self,
        state: Any,
        domain: Optional[str] = None,
        tags: Optional[Set[str]] = None,
    ) -> List[Skill]:
        """
        Retrieve skills applicable to the current state.

        Args:
            state: Current state
            domain: Optional domain filter
            tags: Optional tag filters

        Returns:
            List of applicable skills
        """
        candidates = set(self._skills.keys())

        if domain and domain in self._domain_skills:
            candidates &= self._domain_skills[domain]

        if tags:
            for tag in tags:
                if tag in self._tag_skills:
                    candidates &= self._tag_skills[tag]

        applicable = []
        for name in candidates:
            skill = self._skills[name]
            if skill.can_execute(state):
                applicable.append(skill)

        return applicable

    def retrieve_similar(
        self,
        query_embedding: np.ndarray,
        top_k: int = 5,
        threshold: float = 0.0,
    ) -> List[Tuple[Skill, float]]:
        """
        Retrieve skills similar to a query embedding.

        Returns list of (skill, similarity) tuples.
        """
        similarities = []

        query_norm = np.linalg.norm(query_embedding)
        if query_norm < 1e-8:
            return []

        query_normalized = query_embedding / query_norm

        for skill in self._skills.values():
            if skill.embedding is None:
                continue

            similarity = float(np.dot(query_normalized, skill.embedding))
            if similarity >= threshold:
                similarities.append((skill, similarity))

        similarities.sort(key=lambda x: x[1], reverse=True)
        return similarities[:top_k]

    def compose_skills(
        self,
        skill_names: List[str],
        composition_type: str = "sequence",
        result_name: Optional[str] = None,
    ) -> Optional[Skill]:
        """
        Compose multiple skills into a new skill.

        Args:
            skill_names: Names of skills to compose
            composition_type: "sequence", "parallel", or "conditional"
            result_name: Name for the composed skill

        Returns:
            Composed skill or None if composition failed
        """
        skills = [self._skills[n] for n in skill_names if n in self._skills]
        if len(skills) < 2:
            return None

        if result_name is None:
            result_name = f"composed_{'_'.join(skill_names)}"

        if composition_type == "sequence":
            composed = self._composer.sequence(skills, result_name)
        elif composition_type == "parallel":
            composed = self._composer.parallel(skills, result_name)
        else:
            return None

        for skill in skills:
            skill.child_skills.append(composed.name)

        self.register_skill(composed)
        return composed

    def transfer_skill(
        self,
        skill_name: str,
        target_domain: str,
        adapter: Optional[Callable[[Skill], Skill]] = None,
    ) -> Optional[Skill]:
        """
        Transfer a skill to a new domain.

        Args:
            skill_name: Name of skill to transfer
            target_domain: Target domain
            adapter: Optional function to adapt the skill

        Returns:
            Transferred skill or None
        """
        source_skill = self._skills.get(skill_name)
        if source_skill is None:
            return None

        if adapter:
            transferred = adapter(source_skill)
        else:
            transferred = Skill(
                name=f"{skill_name}_{target_domain}",
                skill_type=source_skill.skill_type,
                preconditions=source_skill.preconditions,
                effects=source_skill.effects,
                policy=source_skill.policy,
                termination=source_skill.termination,
                parameters=dict(source_skill.parameters),
                embedding=source_skill.embedding.copy()
                if source_skill.embedding is not None
                else None,
                parent_skills=[skill_name],
            )

        transferred.metadata.domains.add(target_domain)
        self.register_skill(transferred)

        return transferred

    def record_skill_use(
        self,
        skill_name: str,
        success: bool,
        steps: int,
        reward: float,
    ) -> None:
        """Record a skill usage for statistics."""
        skill = self._skills.get(skill_name)
        if skill:
            skill.metadata.record_use(success, steps, reward, self._current_time)

    def get_best_skill(
        self,
        state: Any,
        metric: str = "success_rate",
    ) -> Optional[Skill]:
        """
        Get the best applicable skill by a metric.

        Metrics: "success_rate", "reward", "efficiency"
        """
        applicable = self.retrieve_applicable(state)
        if not applicable:
            return None

        def get_score(skill: Skill) -> float:
            meta = skill.metadata
            if metric == "success_rate":
                return meta.success_rate
            elif metric == "reward":
                return meta.average_reward
            elif metric == "efficiency":
                if meta.average_steps == 0:
                    return 0.0
                return meta.average_reward / meta.average_steps
            return 0.0

        return max(applicable, key=get_score)

    def advance_time(self, delta: float) -> None:
        """Advance the library's internal clock."""
        self._current_time += delta

    def list_skills(
        self,
        domain: Optional[str] = None,
        skill_type: Optional[SkillType] = None,
    ) -> List[str]:
        """List skill names with optional filters."""
        result = []

        for name, skill in self._skills.items():
            if domain and domain not in skill.metadata.domains:
                continue
            if skill_type and skill.skill_type != skill_type:
                continue
            result.append(name)

        return result

    def statistics(self) -> Dict[str, Any]:
        """Get library statistics."""
        type_counts = {}
        for skill in self._skills.values():
            t = skill.skill_type.value
            type_counts[t] = type_counts.get(t, 0) + 1

        return {
            "total_skills": len(self._skills),
            "domains": len(self._domain_skills),
            "tags": len(self._tag_skills),
            "by_type": type_counts,
            "current_time": self._current_time,
        }



"""
Intrinsic Motivation: Action-State Path Entropy Maximization

Implements the fundamental insight that humans are NOT primarily reward-maximizers.
Instead, humans have an "irreducible desire to act and to move" - maximizing
action-state path entropy to expand possibilities.

Evidence for this view:
- Newborn motor babbling (purposeless movement without reward)
- Children choosing harder games over easy success
- Information seeking even when useless
- Post-satiation exploration (continuing after needs met)

Key properties of path entropy maximization:
- Stochastic behavior (complex rather than rigid patterns)
- Resource-seeking (enables future diversity)
- Risk avoidance (protects possibility space)
- Hierarchical goals (above temporary drives)
"""

import numpy as np
from typing import List, Dict, Optional, Tuple
from dataclasses import dataclass
from collections import deque
from enum import Enum


class DriveType(Enum):
    """Types of intrinsic drives"""

    EXPLORATION = "exploration"  # Seek new states
    MASTERY = "mastery"  # Expand action repertoire
    AUTONOMY = "autonomy"  # Maintain control/options
    CHALLENGE = "challenge"  # Seek optimal difficulty


@dataclass
class ActionState:
    """A point in action-state space"""

    state: np.ndarray
    action: Optional[np.ndarray] = None
    timestamp: float = 0.0
    novelty: float = 0.0
    reachability: float = 1.0  # How accessible from current position


@dataclass
class PossibilityMetrics:
    """Metrics about the current possibility space"""

    volume: float  # Size of reachable space
    diversity: float  # Entropy of visited states
    expansion_rate: float  # How fast space is growing
    constraint_level: float  # How constrained we are
    future_options: int  # Estimated future possibilities


class PossibilitySpace:
    """Represents the space of possible action-state trajectories.

    The core insight: organisms seek to maximize the volume and diversity
    of their possibility space, not just immediate rewards.
    """

    def __init__(
        self,
        state_dim: int,
        action_dim: int,
        history_length: int = 1000,
        expansion_weight: float = 1.0,
    ):
        self.state_dim = state_dim
        self.action_dim = action_dim
        self.history_length = history_length
        self.expansion_weight = expansion_weight

        # Track visited states
        self.visited_states: deque = deque(maxlen=history_length)
        self.visited_actions: deque = deque(maxlen=history_length)

        # Estimate of reachable space boundaries
        self.state_min = np.full(state_dim, np.inf)
        self.state_max = np.full(state_dim, -np.inf)

        # Density estimation for diversity
        self.state_bins = 20
        self.state_histogram = np.zeros([self.state_bins] * min(state_dim, 3))

        # Current position
        self.current_state = np.zeros(state_dim)

    def observe(self, state: np.ndarray, action: Optional[np.ndarray] = None) -> None:
        """Record a visited state-action pair."""
        self.visited_states.append(state.copy())
        if action is not None:
            self.visited_actions.append(action.copy())

        # Update boundaries
        self.state_min = np.minimum(self.state_min, state)
        self.state_max = np.maximum(self.state_max, state)

        # Update histogram for diversity estimation
        self._update_histogram(state)

        self.current_state = state

    def _update_histogram(self, state: np.ndarray) -> None:
        """Update state space histogram for diversity estimation."""
        if len(self.visited_states) < 2:
            return

        # Normalize state to bin indices
        range_vec = self.state_max - self.state_min + 1e-8
        normalized = (state - self.state_min) / range_vec
        normalized = np.clip(normalized, 0, 0.999)

        # Use first 3 dimensions for histogram
        dims = min(self.state_dim, 3)
        indices = tuple((normalized[:dims] * self.state_bins).astype(int))

        try:
            self.state_histogram[indices] += 1
        except IndexError:
            pass

    def compute_volume(self) -> float:
        """Compute volume of explored possibility space."""
        if len(self.visited_states) < 2:
            return 0.0

        range_vec = self.state_max - self.state_min
        range_vec = np.maximum(range_vec, 1e-8)

        # Volume as product of ranges
        volume = np.prod(range_vec)

        # Scale by number of samples
        coverage = len(self.visited_states) / self.history_length

        return volume * coverage

    def compute_diversity(self) -> float:
        """Compute entropy of state distribution (diversity of experiences)."""
        if len(self.visited_states) < 10:
            return 0.0

        # Normalize histogram to probability distribution
        total = np.sum(self.state_histogram)
        if total == 0:
            return 0.0

        probs = self.state_histogram.flatten() / total
        probs = probs[probs > 0]  # Remove zeros for log

        # Shannon entropy
        entropy = -np.sum(probs * np.log(probs + 1e-10))

        # Normalize by max possible entropy
        max_entropy = np.log(len(probs))
        if max_entropy > 0:
            entropy /= max_entropy

        return entropy

    def compute_expansion_rate(self, window: int = 50) -> float:
        """Compute how fast the possibility space is expanding."""
        if len(self.visited_states) < window * 2:
            return 0.0

        states = list(self.visited_states)

        # Compare recent vs older volume
        old_states = np.array(states[-window * 2 : -window])
        new_states = np.array(states[-window:])

        old_volume = np.prod(np.ptp(old_states, axis=0) + 1e-8)
        new_volume = np.prod(np.ptp(new_states, axis=0) + 1e-8)

        expansion = (new_volume - old_volume) / (old_volume + 1e-8)
        return np.clip(expansion, -1.0, 1.0)

    def estimate_future_options(self, horizon: int = 10) -> int:
        """Estimate number of distinct states reachable in horizon steps."""
        if len(self.visited_states) < 20:
            return horizon

        # Estimate transition spread from history
        states = np.array(list(self.visited_states)[-100:])
        if len(states) < 2:
            return horizon

        transitions = np.diff(states, axis=0)
        transition_std = np.std(transitions, axis=0)

        # Estimated reachable volume grows with sqrt(horizon)
        reachable_radius = transition_std * np.sqrt(horizon)
        reachable_volume = np.prod(2 * reachable_radius + 1e-8)

        # Discretize to count of states
        cell_size = np.mean(transition_std) + 1e-8
        n_options = int(reachable_volume / (cell_size**self.state_dim))

        return max(1, min(n_options, 10000))

    def get_metrics(self) -> PossibilityMetrics:
        """Get comprehensive metrics about possibility space."""
        return PossibilityMetrics(
            volume=self.compute_volume(),
            diversity=self.compute_diversity(),
            expansion_rate=self.compute_expansion_rate(),
            constraint_level=1.0 - self.compute_diversity(),
            future_options=self.estimate_future_options(),
        )

    def compute_action_value(self, proposed_action: np.ndarray) -> float:
        """Compute intrinsic value of an action based on possibility expansion.

        Actions that expand the possibility space have higher value.
        """
        if len(self.visited_states) < 10:
            return 0.0

        # Estimate resulting state
        states = np.array(list(self.visited_states)[-20:])
        actions = list(self.visited_actions)[-20:] if self.visited_actions else None

        if actions and len(actions) >= 10:
            # Learn rough action->state_change mapping
            np.array(actions[-10:])
            state_changes = np.diff(states[-11:], axis=0)

            # Simple linear estimate
            mean_change = np.mean(state_changes, axis=0)
            predicted_state = self.current_state + mean_change
        else:
            predicted_state = self.current_state + np.random.randn(self.state_dim) * 0.1

        # Value = how much this expands possibility space
        current_diversity = self.compute_diversity()

        # Temporarily add predicted state
        self.visited_states.append(predicted_state)
        self._update_histogram(predicted_state)
        new_diversity = self.compute_diversity()

        # Remove temporary addition
        self.visited_states.pop()

        expansion_value = new_diversity - current_diversity
        novelty_value = self._compute_novelty(predicted_state)

        return self.expansion_weight * expansion_value + 0.5 * novelty_value

    def _compute_novelty(self, state: np.ndarray) -> float:
        """Compute novelty of a state based on distance from visited states."""
        if len(self.visited_states) < 5:
            return 1.0

        states = np.array(list(self.visited_states)[-100:])
        distances = np.linalg.norm(states - state, axis=1)

        min_dist = np.min(distances)
        mean_dist = np.mean(distances)

        # Novelty high when far from all visited states
        novelty = min_dist / (mean_dist + 1e-8)
        return np.clip(novelty, 0, 1)

    def reset(self) -> None:
        """Reset the possibility space tracking."""
        self.visited_states.clear()
        self.visited_actions.clear()
        self.state_min = np.full(self.state_dim, np.inf)
        self.state_max = np.full(self.state_dim, -np.inf)
        self.state_histogram = np.zeros_like(self.state_histogram)


class ActionDiversityTracker:
    """Tracks and encourages diversity in action selection.

    Implements the observation that organisms exhibit "stochastic behavior"
    with complex rather than rigid patterns.
    """

    def __init__(self, action_dim: int, history_length: int = 500, diversity_target: float = 0.7):
        self.action_dim = action_dim
        self.history_length = history_length
        self.diversity_target = diversity_target

        self.action_history: deque = deque(maxlen=history_length)
        self.action_counts = {}  # Discretized action counts

        # For continuous actions
        self.n_bins = 10

    def record_action(self, action: np.ndarray) -> None:
        """Record an action taken."""
        self.action_history.append(action.copy())

        # Discretize for counting
        key = self._discretize(action)
        self.action_counts[key] = self.action_counts.get(key, 0) + 1

    def _discretize(self, action: np.ndarray) -> tuple:
        """Discretize continuous action for counting."""
        # Assume actions in [-1, 1] range
        bins = ((action + 1) / 2 * self.n_bins).astype(int)
        bins = np.clip(bins, 0, self.n_bins - 1)
        return tuple(bins)

    def compute_diversity(self) -> float:
        """Compute current action diversity (entropy)."""
        if len(self.action_history) < 10:
            return 1.0

        total = sum(self.action_counts.values())
        if total == 0:
            return 1.0

        probs = np.array(list(self.action_counts.values())) / total
        entropy = -np.sum(probs * np.log(probs + 1e-10))

        # Normalize by max entropy
        max_entropy = np.log(float(self.n_bins**self.action_dim))
        return entropy / max_entropy if max_entropy > 0 else 0.0

    def get_diversity_bonus(self, proposed_action: np.ndarray) -> float:
        """Get bonus for taking a less-frequently-used action."""
        if len(self.action_history) < 10:
            return 0.0

        key = self._discretize(proposed_action)
        count = self.action_counts.get(key, 0)
        total = len(self.action_history)

        # Bonus inversely proportional to how often we've taken this action
        frequency = count / total

        # High bonus for rare actions
        bonus = 1.0 - frequency

        return bonus

    def suggest_exploration_direction(self) -> np.ndarray:
        """Suggest an action direction that increases diversity."""
        if len(self.action_history) < 10:
            return np.random.randn(self.action_dim)

        # Find least-used action region
        min_count = float("inf")
        min_key = None

        # Sample random keys and find underexplored region
        for _ in range(20):
            random_action = np.random.rand(self.action_dim) * 2 - 1
            key = self._discretize(random_action)
            count = self.action_counts.get(key, 0)
            if count < min_count:
                min_count = count
                min_key = key

        if min_key is None:
            return np.random.randn(self.action_dim)

        # Convert back to continuous action
        exploration_target = (np.array(min_key) + 0.5) / self.n_bins * 2 - 1
        return exploration_target

    def reset(self) -> None:
        """Reset tracking."""
        self.action_history.clear()
        self.action_counts = {}


class IntrinsicDrive:
    """Represents a specific intrinsic drive (exploration, mastery, etc.).

    Each drive contributes to the overall motivation signal.
    """

    def __init__(
        self,
        drive_type: DriveType,
        base_strength: float = 1.0,
        satiation_rate: float = 0.1,
        recovery_rate: float = 0.05,
    ):
        self.drive_type = drive_type
        self.base_strength = base_strength
        self.satiation_rate = satiation_rate
        self.recovery_rate = recovery_rate

        # Current drive level (0 = satiated, 1 = maximum drive)
        self.level = 0.5

        # History of satisfaction events
        self.satisfaction_history: List[float] = []

    def update(self, satisfaction: float, dt: float = 1.0) -> float:
        """Update drive level based on satisfaction received.

        Args:
            satisfaction: How much this drive was satisfied (0-1)
            dt: Time step

        Returns:
            New drive level
        """
        # Satisfaction reduces drive
        self.level -= self.satiation_rate * satisfaction * dt

        # Natural recovery (drive increases over time without satisfaction)
        self.level += self.recovery_rate * (1.0 - self.level) * dt

        self.level = np.clip(self.level, 0.0, 1.0)
        self.satisfaction_history.append(satisfaction)

        return self.level

    def get_motivation(self) -> float:
        """Get current motivation from this drive."""
        # Motivation is drive level scaled by base strength
        return self.level * self.base_strength

    def get_deprivation(self) -> float:
        """Get deprivation level (how long since satisfaction)."""
        if not self.satisfaction_history:
            return 1.0

        recent = self.satisfaction_history[-20:]
        mean_satisfaction = np.mean(recent)
        return 1.0 - mean_satisfaction


class PathEntropyMaximizer:
    """The core intrinsic motivation system.

    Implements the principle that organisms maximize action-state path entropy
    rather than cumulative reward. This explains:
    - Exploration without immediate reward
    - Preference for having options
    - Resource-seeking (enables future possibilities)
    - Risk avoidance (protects possibility space)
    """

    def __init__(
        self,
        state_dim: int,
        action_dim: int,
        entropy_weight: float = 1.0,
        diversity_weight: float = 0.5,
        risk_aversion: float = 0.3,
    ):
        self.state_dim = state_dim
        self.action_dim = action_dim
        self.entropy_weight = entropy_weight
        self.diversity_weight = diversity_weight
        self.risk_aversion = risk_aversion

        # Core components
        self.possibility_space = PossibilitySpace(state_dim, action_dim)
        self.action_tracker = ActionDiversityTracker(action_dim)

        # Intrinsic drives
        self.drives = {
            DriveType.EXPLORATION: IntrinsicDrive(DriveType.EXPLORATION, 1.0),
            DriveType.MASTERY: IntrinsicDrive(DriveType.MASTERY, 0.8),
            DriveType.AUTONOMY: IntrinsicDrive(DriveType.AUTONOMY, 1.2),
            DriveType.CHALLENGE: IntrinsicDrive(DriveType.CHALLENGE, 0.6),
        }

        # Track path entropy
        self.path_history: List[Tuple[np.ndarray, np.ndarray]] = []
        self.max_path_length = 1000

    def observe(self, state: np.ndarray, action: np.ndarray, extrinsic_reward: float = 0.0) -> None:
        """Observe a state-action transition."""
        self.possibility_space.observe(state, action)
        self.action_tracker.record_action(action)

        self.path_history.append((state.copy(), action.copy()))
        if len(self.path_history) > self.max_path_length:
            self.path_history.pop(0)

        # Update drives based on what was satisfied
        self._update_drives(state, action, extrinsic_reward)

    def _update_drives(
        self, state: np.ndarray, action: np.ndarray, extrinsic_reward: float
    ) -> None:
        """Update intrinsic drives based on transition."""
        # Exploration: satisfied by novelty
        novelty = self.possibility_space._compute_novelty(state)
        self.drives[DriveType.EXPLORATION].update(novelty)

        # Mastery: satisfied by action diversity
        diversity = self.action_tracker.compute_diversity()
        self.drives[DriveType.MASTERY].update(diversity)

        # Autonomy: satisfied by expansion of possibilities
        metrics = self.possibility_space.get_metrics()
        expansion = max(0, metrics.expansion_rate)
        self.drives[DriveType.AUTONOMY].update(expansion)

        # Challenge: satisfied by optimal difficulty (not too easy, not too hard)
        # Approximate with distance traveled
        if len(self.path_history) >= 2:
            distance = np.linalg.norm(state - self.path_history[-2][0])
            optimal_distance = 0.3  # Target difficulty
            challenge_satisfaction = 1.0 - abs(distance - optimal_distance) / optimal_distance
            challenge_satisfaction = np.clip(challenge_satisfaction, 0, 1)
        else:
            challenge_satisfaction = 0.5
        self.drives[DriveType.CHALLENGE].update(challenge_satisfaction)

    def compute_intrinsic_motivation(self) -> float:
        """Compute total intrinsic motivation level."""
        total = 0.0
        for drive in self.drives.values():
            total += drive.get_motivation()
        return total / len(self.drives)

    def compute_action_value(
        self, proposed_action: np.ndarray, expected_state: Optional[np.ndarray] = None
    ) -> float:
        """Compute intrinsic value of a proposed action.

        Value comes from:
        1. Path entropy expansion
        2. Action diversity
        3. Possibility space expansion
        4. Risk assessment
        """
        value = 0.0

        # Path entropy contribution
        path_entropy_value = self.possibility_space.compute_action_value(proposed_action)
        value += self.entropy_weight * path_entropy_value

        # Action diversity contribution
        diversity_bonus = self.action_tracker.get_diversity_bonus(proposed_action)
        value += self.diversity_weight * diversity_bonus

        # Drive-weighted contributions
        exploration_drive = self.drives[DriveType.EXPLORATION].get_motivation()
        value += 0.3 * exploration_drive * diversity_bonus

        # Risk penalty (avoid actions that might constrain future possibilities)
        if expected_state is not None:
            risk = self._assess_risk(expected_state)
            value -= self.risk_aversion * risk

        return value

    def _assess_risk(self, expected_state: np.ndarray) -> float:
        """Assess risk that a state constrains future possibilities."""
        metrics = self.possibility_space.get_metrics()

        # Risk higher near boundaries of explored space
        if len(self.possibility_space.visited_states) < 10:
            return 0.0

        # Distance from center of explored space
        states = np.array(list(self.possibility_space.visited_states))
        center = np.mean(states, axis=0)
        max_dist = np.max(np.linalg.norm(states - center, axis=1))

        dist_from_center = np.linalg.norm(expected_state - center)
        relative_dist = dist_from_center / (max_dist + 1e-8)

        # Risk high at edges (might lead to dead ends)
        if relative_dist > 0.9:
            return 0.5

        # Risk also high if space is already constrained
        if metrics.constraint_level > 0.7:
            return 0.3

        return 0.0

    def compute_path_entropy(self, horizon: int = 50) -> float:
        """Compute entropy of recent path through state-action space."""
        if len(self.path_history) < horizon:
            return 0.0

        recent_path = self.path_history[-horizon:]

        # State entropy
        states = np.array([p[0] for p in recent_path])
        state_cov = np.cov(states.T)
        if np.ndim(state_cov) == 0:
            state_cov = np.array([[state_cov]])
        state_entropy = 0.5 * np.log(
            np.linalg.det(state_cov + 1e-8 * np.eye(self.state_dim)) + 1e-10
        )

        # Action entropy
        actions = np.array([p[1] for p in recent_path])
        action_cov = np.cov(actions.T)
        if np.ndim(action_cov) == 0:
            action_cov = np.array([[action_cov]])
        action_entropy = 0.5 * np.log(
            np.linalg.det(action_cov + 1e-8 * np.eye(self.action_dim)) + 1e-10
        )

        return state_entropy + action_entropy

    def suggest_action(self, current_state: np.ndarray) -> np.ndarray:
        """Suggest an action that maximizes intrinsic motivation."""
        # Sample candidate actions
        n_candidates = 20
        candidates = [np.random.randn(self.action_dim) * 0.5 for _ in range(n_candidates)]

        # Add diversity-seeking direction
        exploration_dir = self.action_tracker.suggest_exploration_direction()
        candidates.append(exploration_dir)

        # Evaluate each
        best_action = candidates[0]
        best_value = float("-inf")

        for action in candidates:
            # Clip to valid range
            action = np.clip(action, -1, 1)
            value = self.compute_action_value(action)
            if value > best_value:
                best_value = value
                best_action = action

        return best_action

    def get_drive_levels(self) -> Dict[DriveType, float]:
        """Get current level of each intrinsic drive."""
        return {dt: drive.level for dt, drive in self.drives.items()}

    def get_metrics(self) -> Dict[str, float]:
        """Get comprehensive metrics about motivation state."""
        metrics = self.possibility_space.get_metrics()
        return {
            "intrinsic_motivation": self.compute_intrinsic_motivation(),
            "path_entropy": self.compute_path_entropy(),
            "possibility_volume": metrics.volume,
            "state_diversity": metrics.diversity,
            "action_diversity": self.action_tracker.compute_diversity(),
            "expansion_rate": metrics.expansion_rate,
            "future_options": metrics.future_options,
            "exploration_drive": self.drives[DriveType.EXPLORATION].level,
            "mastery_drive": self.drives[DriveType.MASTERY].level,
            "autonomy_drive": self.drives[DriveType.AUTONOMY].level,
            "challenge_drive": self.drives[DriveType.CHALLENGE].level,
        }

    def reset(self) -> None:
        """Reset the motivation system."""
        self.possibility_space.reset()
        self.action_tracker.reset()
        self.path_history = []
        for drive in self.drives.values():
            drive.level = 0.5
            drive.satisfaction_history = []



"""
neuro-inference: Probabilistic and causal reasoning.

This module provides:
- Bayesian networks and belief propagation
- Causal models and intervention reasoning
- Analogical reasoning and structure mapping
"""

from .bayesian.networks import (
    BayesianNetwork,
    DiscreteNode,
    ContinuousNode,
    CPT,
    NetworkQuery,
)
from .bayesian.belief_prop import (
    BeliefPropagation,
    Message,
    FactorGraph,
    Factor,
)
from .bayesian.learning import (
    StructureLearner,
    ParameterLearner,
    BayesianScore,
)
from .causal.scm import (
    StructuralCausalModel,
    CausalVariable,
    StructuralEquation,
)
from .causal.intervention import (
    Intervention,
    InterventionEngine,
    DoOperator,
)
from .causal.counterfactual import (
    CounterfactualReasoner,
    CounterfactualQuery,
    PotentialOutcome,
)
from .analogical.mapping import (
    StructureMapper,
    Analogy,
    StructuralAlignment,
)
from .analogical.retrieval import (
    AnalogyRetriever,
    CaseLibrary,
    Case,
)
from .integration import (
    ProbabilisticReasoner,
    InferenceResult,
)

__all__ = [
    # Bayesian
    "BayesianNetwork",
    "DiscreteNode",
    "ContinuousNode",
    "CPT",
    "NetworkQuery",
    "BeliefPropagation",
    "Message",
    "FactorGraph",
    "Factor",
    "StructureLearner",
    "ParameterLearner",
    "BayesianScore",
    # Causal
    "StructuralCausalModel",
    "CausalVariable",
    "StructuralEquation",
    "Intervention",
    "InterventionEngine",
    "DoOperator",
    "CounterfactualReasoner",
    "CounterfactualQuery",
    "PotentialOutcome",
    # Analogical
    "StructureMapper",
    "Analogy",
    "StructuralAlignment",
    "AnalogyRetriever",
    "CaseLibrary",
    "Case",
    # Integration
    "ProbabilisticReasoner",
    "InferenceResult",
]



"""
Neuro AGI Demo Runner.

Entry point for `neuro-demo` CLI command.

Usage:
    neuro-demo all        # Run all demos
    neuro-demo medical    # Medical diagnosis reasoning
    neuro-demo physics    # Physics concept learning
    neuro-demo language   # Language understanding
    neuro-demo arc        # ARC-style pattern reasoning
    neuro-demo continual  # Continual learning demo
"""

import argparse
import sys
from pathlib import Path


def get_version():
    """Get version from __init__.py."""
    return "0.9.0"


def run_medical():
    """Run medical diagnosis demo."""
    import sys
    from pathlib import Path

    # Ensure demos directory is in path first
    demos_dir = Path(__file__).parent.parent / "demos" / "agi_integration"
    sys.path.insert(0, str(demos_dir))
    try:
        from run_demo import demo_medical_diagnosis

        demo_medical_diagnosis()
    finally:
        sys.path.remove(str(demos_dir))


def run_physics():
    """Run physics learning demo."""
    import sys
    from pathlib import Path

    demos_dir = Path(__file__).parent.parent / "demos" / "agi_integration"
    sys.path.insert(0, str(demos_dir))
    try:
        from run_demo import demo_physics_learning

        demo_physics_learning()
    finally:
        sys.path.remove(str(demos_dir))


def run_language():
    """Run language understanding demo."""
    import sys
    from pathlib import Path

    demos_dir = Path(__file__).parent.parent / "demos" / "agi_integration"
    sys.path.insert(0, str(demos_dir))
    try:
        from run_demo import demo_language_concepts

        demo_language_concepts()
    finally:
        sys.path.remove(str(demos_dir))


def run_arc():
    """Run ARC-style reasoning demo."""
    import sys
    from pathlib import Path

    demos_dir = Path(__file__).parent.parent / "demos" / "agi_integration"
    sys.path.insert(0, str(demos_dir))
    try:
        from scenario_arc import demo_arc_reasoning

        demo_arc_reasoning()
    except ImportError as e:
        print(f"ARC demo import error: {e}")
        print("Make sure neuro-abstract, neuro-causal, neuro-meta-reasoning are available.")
        return 1
    finally:
        if str(demos_dir) in sys.path:
            sys.path.remove(str(demos_dir))
    return 0


def run_continual():
    """Run continual learning demo."""
    import sys
    from pathlib import Path

    demos_dir = Path(__file__).parent.parent / "demos" / "agi_integration"
    sys.path.insert(0, str(demos_dir))
    try:
        from scenario_continual import demo_continual_learning

        demo_continual_learning()
    except ImportError as e:
        print(f"Continual learning demo import error: {e}")
        print("Make sure neuro-continual is available.")
        return 1
    finally:
        if str(demos_dir) in sys.path:
            sys.path.remove(str(demos_dir))
    return 0


def run_all():
    """Run all demos."""
    print("\n" + "#" * 60)
    print("#" + " " * 18 + "NEURO AGI DEMOS" + " " * 19 + "#")
    print("#" * 60)
    print(f"\nVersion: {get_version()}")
    print("Running all available demos...\n")

    run_medical()
    run_physics()
    run_language()
    run_arc()
    run_continual()

    print("\n" + "=" * 60)
    print("All demos complete!")
    print("=" * 60 + "\n")


def list_demos():
    """List available demos."""
    print("Available demos:")
    print("  medical   - Medical diagnosis reasoning")
    print("  physics   - Physics concept learning")
    print("  language  - Language understanding")
    print("  arc       - ARC-style pattern reasoning")
    print("  continual - Continual learning demo")
    print("  all       - Run all 5 demos")


def main():
    """Main entry point for neuro-demo CLI."""
    parser = argparse.ArgumentParser(
        prog="neuro-demo",
        description="Neuro AGI Demo Runner - Showcase cognitive capabilities",
    )
    parser.add_argument(
        "-v",
        "--version",
        action="store_true",
        help="Show version and exit",
    )
    parser.add_argument(
        "-l",
        "--list",
        action="store_true",
        help="List available demos",
    )
    parser.add_argument(
        "demo",
        nargs="?",
        choices=["medical", "physics", "language", "arc", "continual", "all"],
        help="Demo to run",
    )

    args = parser.parse_args()

    if args.version:
        print(f"neuro-demo {get_version()}")
        return 0

    if args.list:
        list_demos()
        return 0

    if args.demo is None:
        parser.print_help()
        print("\nRun 'neuro-demo --list' to see available demos.")
        return 0

    # Add project root to path for imports
    project_root = Path(__file__).parent.parent
    if str(project_root) not in sys.path:
        sys.path.insert(0, str(project_root))

    demos = {
        "medical": run_medical,
        "physics": run_physics,
        "language": run_language,
        "arc": run_arc,
        "continual": run_continual,
        "all": run_all,
    }

    return demos[args.demo]() or 0


if __name__ == "__main__":
    sys.exit(main())



"""
Quantization: Reduce precision for efficiency.

Implements various quantization methods for model compression.
"""

from dataclasses import dataclass
from typing import Dict, List, Optional, Any, Tuple
from enum import Enum
import numpy as np


class QuantizationLevel(Enum):
    """Quantization precision levels."""

    FP32 = 32  # Full precision
    FP16 = 16  # Half precision
    INT8 = 8  # 8-bit integer
    INT4 = 4  # 4-bit integer
    BINARY = 1  # Binary


@dataclass
class QuantizationConfig:
    """Configuration for quantization."""

    level: QuantizationLevel = QuantizationLevel.INT8
    symmetric: bool = True
    per_channel: bool = False
    calibration_samples: int = 100


@dataclass
class QuantizedTensor:
    """A quantized tensor with scale and zero-point."""

    data: np.ndarray  # Quantized values
    scale: np.ndarray  # Scale factor(s)
    zero_point: np.ndarray  # Zero point(s)
    original_dtype: np.dtype
    level: QuantizationLevel

    def dequantize(self) -> np.ndarray:
        """Convert back to original precision."""
        return (self.data.astype(np.float32) - self.zero_point) * self.scale


@dataclass
class QuantizedModel:
    """A quantized model."""

    weights: Dict[str, QuantizedTensor]
    level: QuantizationLevel
    memory_reduction: float
    accuracy_change: Optional[float] = None


class Quantizer:
    """
    Quantizer for neural network weights and activations.

    Implements:
    - Symmetric and asymmetric quantization
    - Per-tensor and per-channel quantization
    - Various bit widths
    """

    def __init__(self, config: Optional[QuantizationConfig] = None):
        self.config = config or QuantizationConfig()

        # Bit ranges
        self._bit_ranges = {
            QuantizationLevel.INT8: (-128, 127),
            QuantizationLevel.INT4: (-8, 7),
            QuantizationLevel.BINARY: (-1, 1),
            QuantizationLevel.FP16: None,  # No range limit
            QuantizationLevel.FP32: None,
        }

    def quantize(
        self,
        tensor: np.ndarray,
        level: Optional[QuantizationLevel] = None,
        symmetric: Optional[bool] = None,
    ) -> QuantizedTensor:
        """
        Quantize a tensor.

        Args:
            tensor: Input tensor (float32)
            level: Target quantization level
            symmetric: Use symmetric quantization

        Returns:
            QuantizedTensor
        """
        level = level or self.config.level
        symmetric = symmetric if symmetric is not None else self.config.symmetric

        # FP16 - just convert
        if level == QuantizationLevel.FP16:
            return QuantizedTensor(
                data=tensor.astype(np.float16),
                scale=np.array([1.0]),
                zero_point=np.array([0.0]),
                original_dtype=tensor.dtype,
                level=level,
            )

        # FP32 - no quantization
        if level == QuantizationLevel.FP32:
            return QuantizedTensor(
                data=tensor.copy(),
                scale=np.array([1.0]),
                zero_point=np.array([0.0]),
                original_dtype=tensor.dtype,
                level=level,
            )

        # Integer quantization
        qmin, qmax = self._bit_ranges[level]

        if self.config.per_channel and tensor.ndim > 1:
            return self._quantize_per_channel(tensor, qmin, qmax, symmetric, level)
        else:
            return self._quantize_per_tensor(tensor, qmin, qmax, symmetric, level)

    def _quantize_per_tensor(
        self,
        tensor: np.ndarray,
        qmin: int,
        qmax: int,
        symmetric: bool,
        level: QuantizationLevel,
    ) -> QuantizedTensor:
        """Quantize entire tensor with single scale/zero-point."""
        if symmetric:
            max_abs = np.max(np.abs(tensor))
            scale = max_abs / max(abs(qmin), abs(qmax))
            zero_point = 0
        else:
            min_val, max_val = tensor.min(), tensor.max()
            scale = (max_val - min_val) / (qmax - qmin)
            zero_point = qmin - np.round(min_val / scale)

        scale = max(scale, 1e-8)  # Prevent divide by zero

        # Quantize
        quantized = np.round(tensor / scale + zero_point)
        quantized = np.clip(quantized, qmin, qmax)

        return QuantizedTensor(
            data=quantized.astype(np.int8 if level == QuantizationLevel.INT8 else np.int8),
            scale=np.array([scale]),
            zero_point=np.array([zero_point]),
            original_dtype=tensor.dtype,
            level=level,
        )

    def _quantize_per_channel(
        self,
        tensor: np.ndarray,
        qmin: int,
        qmax: int,
        symmetric: bool,
        level: QuantizationLevel,
    ) -> QuantizedTensor:
        """Quantize with per-channel scale/zero-point."""
        n_channels = tensor.shape[0]
        scales = np.zeros(n_channels)
        zero_points = np.zeros(n_channels)
        quantized = np.zeros_like(tensor, dtype=np.int8)

        for c in range(n_channels):
            channel_data = tensor[c]

            if symmetric:
                max_abs = np.max(np.abs(channel_data))
                scale = max_abs / max(abs(qmin), abs(qmax))
                zp = 0
            else:
                min_val, max_val = channel_data.min(), channel_data.max()
                scale = (max_val - min_val) / (qmax - qmin)
                zp = qmin - np.round(min_val / scale)

            scale = max(scale, 1e-8)
            scales[c] = scale
            zero_points[c] = zp

            # Quantize channel
            q = np.round(channel_data / scale + zp)
            quantized[c] = np.clip(q, qmin, qmax)

        return QuantizedTensor(
            data=quantized,
            scale=scales,
            zero_point=zero_points,
            original_dtype=tensor.dtype,
            level=level,
        )

    def dequantize(self, qtensor: QuantizedTensor) -> np.ndarray:
        """Dequantize a tensor."""
        return qtensor.dequantize()

    def quantize_model(
        self,
        weights: Dict[str, np.ndarray],
        level: Optional[QuantizationLevel] = None,
    ) -> QuantizedModel:
        """
        Quantize all weights in a model.

        Args:
            weights: Dictionary of layer_name -> weight tensor
            level: Target quantization level

        Returns:
            QuantizedModel
        """
        level = level or self.config.level

        quantized_weights = {}
        original_size = 0
        quantized_size = 0

        for name, weight in weights.items():
            original_size += weight.nbytes

            qtensor = self.quantize(weight, level)
            quantized_weights[name] = qtensor

            quantized_size += qtensor.data.nbytes + qtensor.scale.nbytes + qtensor.zero_point.nbytes

        memory_reduction = 1.0 - quantized_size / original_size if original_size > 0 else 0.0

        return QuantizedModel(
            weights=quantized_weights,
            level=level,
            memory_reduction=memory_reduction,
        )

    def dequantize_model(
        self,
        qmodel: QuantizedModel,
    ) -> Dict[str, np.ndarray]:
        """Dequantize all weights in a model."""
        return {name: qtensor.dequantize() for name, qtensor in qmodel.weights.items()}

    def compute_quantization_error(
        self,
        original: np.ndarray,
        quantized: QuantizedTensor,
    ) -> Dict[str, float]:
        """Compute quantization error metrics."""
        dequantized = quantized.dequantize()

        mse = np.mean((original - dequantized) ** 2)
        mae = np.mean(np.abs(original - dequantized))
        max_error = np.max(np.abs(original - dequantized))

        # Signal-to-noise ratio
        signal_power = np.mean(original**2)
        noise_power = mse
        snr_db = 10 * np.log10(signal_power / (noise_power + 1e-10))

        return {
            "mse": float(mse),
            "mae": float(mae),
            "max_error": float(max_error),
            "snr_db": float(snr_db),
        }

    def calibrate(
        self,
        tensor: np.ndarray,
        calibration_data: List[np.ndarray],
    ) -> Tuple[float, float]:
        """
        Calibrate scale and zero-point using data.

        Args:
            tensor: Weight tensor
            calibration_data: Activation samples for calibration

        Returns:
            Tuple of (scale, zero_point)
        """
        # Collect activation statistics
        all_vals = []
        for data in calibration_data[: self.config.calibration_samples]:
            all_vals.extend(data.flatten())

        all_vals = np.array(all_vals)

        # Use percentile to handle outliers
        min_val = np.percentile(all_vals, 0.1)
        max_val = np.percentile(all_vals, 99.9)

        qmin, qmax = self._bit_ranges[self.config.level]

        if self.config.symmetric:
            max_abs = max(abs(min_val), abs(max_val))
            scale = max_abs / max(abs(qmin), abs(qmax))
            zero_point = 0.0
        else:
            scale = (max_val - min_val) / (qmax - qmin)
            zero_point = qmin - min_val / scale

        return float(scale), float(zero_point)

    def statistics(self) -> Dict[str, Any]:
        """Get quantizer statistics."""
        return {
            "level": self.config.level.value,
            "bits": self.config.level.value,
            "symmetric": self.config.symmetric,
            "per_channel": self.config.per_channel,
        }
